/*
 * This file is part of firestorm NIDS
 * Copyright (c) 2003,2004 Gianni Tedesco
 * Released under the terms of the GNU GPL version 2
 *
 * This a special capdev which is a test rig for TCP application layer
 * decoders. It takes 2 files one for each side of the TCP stream and
 * randomly packetizes them and injects them in to a specific decoder.
*/
#include <firestorm.h>

#include <sys/socket.h>
#include <netinet/in.h>
#include <arpa/inet.h>
#include <unistd.h>

#include <plugin.h>
#include <args.h>
#include <alert.h>
#include <signature.h>
#include <decode.h>
#include <capture.h>
#include <args.h>

static struct proto *ip_proto;
static struct proto *tcp_proto;

#define PKT_HDR_LEN (sizeof(struct pkt_iphdr) + sizeof(struct pkt_tcphdr))
static size_t fag_minstream=128;
static size_t fag_maxstream=4096;

struct fag_side {
	FILE *fp;
	u_int16_t port;
	u_int32_t addr;

	int queue;
	void *buf;
	size_t buf_ofs;
	size_t buf_sz;
};

/* This is our own private data */
struct fag_priv {
	struct packet pkt;
	struct fag_side c;
	struct fag_side s;
	size_t maxseg;
	char *proto;
	struct proto *p;
	char *fbuf;
};

static int cb_server(struct arg *a, void *priv)
{
	struct fag_priv *p=priv;

	if ( p == NULL )
		return 0;

	if ( p->s.fp )
		mesg(M_WARN, "fagrouter: you can only have one server file");

	p->s.fp = fopen(a->val.v_str, "r");
	if ( p->s.fp == NULL ) {
		mesg(M_ERR, "fagrouter: %s: fopen(): %s", a->val.v_str, sys_err());
		return 0;
	}

	return 1;
}

static int cb_client(struct arg *a, void *priv)
{
	struct fag_priv *p=priv;

	if ( p == NULL )
		return 0;

	if ( p->c.fp )
		mesg(M_WARN, "fagrouter: you can only have one client file");

	p->c.fp = fopen(a->val.v_str, "r");
	if ( p->c.fp == NULL ) {
		mesg(M_ERR, "fagrouter: %s: fopen(): %s", a->val.v_str, sys_err());
		return 0;
	}
	return 1;
}

static int cb_proto(struct arg *a, void *priv)
{
	struct fag_priv *p=priv;

	if ( p == NULL )
		return 0;

	if ( p->proto )
		mesg(M_WARN, "fagrouter: you can only have one protocol");

	p->proto = strdup(a->val.v_str);
	if ( p->proto == NULL ) {
		mesg(M_ERR, "fagrouter: strdup(): %s", sys_err());
		return 0;
	}

	return 1;
}

/* Stop the capture and cleanup our process */
static void fag_end(void *priv)
{
	struct fag_priv *p=priv;

	if ( p == NULL )
		return;

	if ( p->s.fp )
		fclose(p->s.fp);
	if ( p->c.fp )
		fclose(p->c.fp);
	if ( p->proto )
		free(p->proto);

	if ( p->p && p->p->flow_dtor && p->fbuf ) {
		p->p->flow_dtor(p->fbuf);
		free(p->fbuf);
	}

	free(p);
}

/* Initialise a capture process, we open the file and then
 * return our opaque private data structure to firestorm */
static int fag_init(const char *args, struct capture *capture, void **priv)
{
	struct fag_priv *p=NULL;
	size_t maxseg = 512;
	struct arg fag_args[]={
		{"server", ARGTYPE_STRING, cb_server},
		{"client", ARGTYPE_STRING, cb_client},
		{"proto",  ARGTYPE_STRING, cb_proto},
		{"maxseg", ARGTYPE_PBYTES, NULL, {.vp_bytes=&maxseg}},
		{NULL,ARGTYPE_NOP,NULL}
	};

	if ( !args )
		return 0;

	if ( !(p=calloc(1, sizeof(*p))) )
		return 0;

	switch( args_parse(fag_args, args, p) ) {
	case -1:
		mesg(M_ERR, "fagrouter: parse error: %s", args);
	case 0: /* fall through */
		goto err;
	default:
		break;
	}

	/* Could select too huge a size and then fill the
	 * stack causing a segfault, not the best behaivour
	 * but I'd rather not impose arbitrary limit */
	if ( (p->maxseg=maxseg) == 0 ) {
		mesg(M_WARN, "fagrouter: maxseg cannot be zero");
		goto err;
	}

	if ( !p->s.fp || !p->c.fp ) {
		mesg(M_ERR, "fagrouter: must specify client AND server");
		goto err;
	}

	capdev_init_packet(&p->pkt, capture);
	*priv = p;
	return 1;
err:
	fag_end(p);
	return 0;
}

static void fag_freebuf(struct fag_side *rcv)
{
	if ( rcv->buf )
		free(rcv->buf);
	rcv->buf = NULL;
	rcv->buf_sz = 0;
	rcv->buf_ofs = 0;
	rcv->queue = 0;
	
}

static inline void do_packet(struct fag_priv *p, struct fag_side *snd,
				struct fag_side *rcv, int len)
{
	char buf[PKT_HDR_LEN + fag_maxstream];
	struct pkt_iphdr *iph;
	struct pkt_tcphdr *tcph;
	struct tcp_session s;
	size_t sz;
	size_t ret;
	char *ptr;
	
	iph = (struct pkt_iphdr *)buf;
	tcph = (struct pkt_tcphdr *)(buf + sizeof(*iph));
	ptr = buf + PKT_HDR_LEN;

#if 1
	sz = fread(ptr, 1, len, snd->fp);
#else
	if ( !fgets(ptr, p->maxseg, snd->fp) )
		return;
	sz = strchr(ptr, '\n') - ptr;
	sz += 1;
#endif
	if ( !sz )
		return;

	/* IP header */
	iph->version = 4;
	iph->ihl = sizeof(*iph)>>2;
	iph->tos = 0;
	iph->tot_len = htons(sz + PKT_HDR_LEN);
	iph->id = 0xfeed;
	iph->frag_off = 0;
	iph->ttl = 0xff;
	iph->protocol = 6;
	iph->csum = 0;
	iph->saddr = snd->addr;
	iph->daddr = rcv->addr;

	/* TCP header */
	tcph->sport = snd->port;
	tcph->dport = rcv->port;
	tcph->seq = 0;
	tcph->ack = 0;
	tcph->res1 = 0;
	tcph->doff = sizeof(*tcph)>>2;
	tcph->flags.flags = TCP_PSH|TCP_ACK;
	tcph->win = htons(8192);
	tcph->csum = 0;
	tcph->urp = 0;

	/* TCP session */
	s.next = s.prev = NULL;
	s.tmo_next = s.tmo_prev = NULL;
	s.hash_next = NULL;
	s.hash_pprev = (struct tcp_session **)NULL;
	s.c_addr = p->c.addr;
	s.s_addr = p->s.addr;
	s.c_port = p->c.port;
	s.s_port = p->s.port;
	s.proto = p->p;
	s.flow = p->fbuf;
	s.bucket = 0;
	s.expire = 0;

	/* Packet */
	p->pkt.len = sz + PKT_HDR_LEN;
	p->pkt.caplen = sz + PKT_HDR_LEN;
	p->pkt.base = buf;
	p->pkt.end = p->pkt.base + p->pkt.caplen;
	p->pkt.alert = NULL;
	gettimeofday(&p->pkt.time, NULL);
	p->pkt.flags = FP_COPY|FP_PROMISC|FP_PKTTYPE|FP_REAL|HOST_ORDER;
	p->pkt.llen = 2;

	serial_number(&p->pkt);

	p->pkt.layer[0].h.ip = iph;
	p->pkt.layer[0].proto = ip_proto;
	p->pkt.layer[0].flags = FLAG_IP_CSUM;
	p->pkt.layer[0].session = NULL;

	p->pkt.layer[1].h.tcp = tcph;
	p->pkt.layer[1].proto = tcp_proto;
	p->pkt.layer[1].flags = FLAG_TCP_CSUM|FLAG_TCP_STATE|FLAG_TCP_SURE|FLAG_TCP_TRACK;
	p->pkt.layer[1].session = &s;
	if ( rcv == &p->s )
		p->pkt.layer[1].flags |= FLAG_TCP_2SVR;

	p->pkt.layer[2].h.raw = ptr;
	p->pkt.layer[2].proto = p->p;
	p->pkt.layer[2].flags = 0;
	p->pkt.layer[2].session = NULL;

	if ( rcv->queue ) {
		size_t len;

		len = rcv->buf_ofs + sz;
		if ( len > rcv->buf_sz ) {
			void *new;

			len = rcv->buf_sz * 2;
			if ( len > fag_maxstream )
				len = fag_maxstream;

			new = realloc(rcv->buf, len);
			if ( new ) {
				rcv->buf = new;
				rcv->buf_sz = len;
			}
		}

		len = rcv->buf_sz - rcv->buf_ofs;
		len = (len > sz) ? sz : len;

		if ( len == 0 ) {
			mesg(M_DEBUG, "fagrouter: reasm too big");
			fag_freebuf(rcv);
			return;
		}

		memcpy(rcv->buf + rcv->buf_ofs, ptr, len);
		rcv->buf_ofs += len;

again:
		ret = p->p->sdecode(&p->pkt, rcv->buf, rcv->buf_ofs);
		if ( ret == 0 )
			return;

		/* Generate reassemblygram */
		memcpy(ptr, rcv->buf, ret);
		p->pkt.len = ret + PKT_HDR_LEN;
		p->pkt.caplen = ret + PKT_HDR_LEN;
		p->pkt.end = p->pkt.base + p->pkt.caplen;
		iph->tot_len = htons(ret + PKT_HDR_LEN);
		p->p->decode(&p->pkt);

		if ( ret >= rcv->buf_ofs ) {
			fag_freebuf(rcv);
			return;
		}

		memmove(rcv->buf, rcv->buf + ret, rcv->buf_ofs - ret);
		rcv->buf_ofs -= ret;
		goto again;
	}else{
		size_t dlen, len;
		void *dptr;

		ret = p->p->sdecode(&p->pkt, ptr, sz);

		if ( ret )
			p->p->decode(&p->pkt);

		if ( ret == sz )
			return;

		/* Skip over the bit that was OK */
		dlen = sz - ret;
		dptr = ptr + ret;

		if ( dlen > fag_maxstream ) {
			mesg(M_DEBUG, "fagrouter: initial segment too big");
			return;
		}

		len = ( dlen > fag_minstream ) ?
			dlen * 2 : fag_minstream;

		if ( len > fag_maxstream )
			len = fag_maxstream;

		rcv->buf = malloc(len);
		if ( !rcv->buf ) {
			mesg(M_DEBUG, "fagrouter: malloc(): %s", sys_err());
			return;
		}

		rcv->queue = 1;
		rcv->buf_ofs = dlen;
		rcv->buf_sz = len;
		memcpy(rcv->buf, dptr, dlen);
	}
}

static int fag_get_port(struct proto *p)
{
	struct proto_child *c;
	int ret = -1;

	for( c=tcp_proto->children; c; c=c->next ) {
		if ( c->proto == p )
			ret = c->id;
	}

	return ret;
}

static void fag_go(void *priv, struct capture *c)
{
	struct fag_priv *p = priv;
	struct proto *proto;
	int port;

	if ( p == NULL )
		return;

	/* Get all the right protocols */
	p->p = proto = decode_proto(p->proto);
	if ( proto == NULL ) {
		c->state = CAP_STATE_STOP;
		mesg(M_ERR, "fagrouter: %s: unknown protocol", p->proto);
		return;
	}

	ip_proto = decode_proto("ip");
	if ( ip_proto == NULL ) {
		c->state = CAP_STATE_STOP;
		mesg(M_ERR, "fagrouter: IP protocol must be installed");
		return;
	}

	tcp_proto = decode_proto("tcp");
	if ( tcp_proto == NULL ) {
		c->state = CAP_STATE_STOP;
		mesg(M_ERR, "fagrouter: TCP protocol must be installed");
		return;
	}

	/* Setup the flow buffer */
	if ( p->fbuf == NULL ) {
		p->fbuf = malloc(proto->flow_len);
		if ( !p->fbuf ) {
			c->state = CAP_STATE_STOP;
			mesg(M_ERR, "fagrouter: %s: not a TCP protocol", p->proto);
			return;
		}

		if ( proto->flow_ctor )
			proto->flow_ctor(p->fbuf);
	}

	port = fag_get_port(proto);
	if ( port < 0 ) {
		c->state = CAP_STATE_STOP;
		mesg(M_ERR, "fagrouter: %s: not a TCP protocol", p->proto);
		return;
	}

	p->s.port = port;
	p->c.port = htons(30000);
	inet_aton("192.168.0.1", (struct in_addr *)&p->s.addr);
	inet_aton("192.168.0.2", (struct in_addr *)&p->c.addr);

	/* Main loop */
	while(c->state == CAP_STATE_CAPTURE) {
		int r = rand();
		int len = ((r>>1) % p->maxseg) + 1;
		int svr = r & 0x01;
		struct fag_side *snd, *rcv;

		if ( feof(p->c.fp) && feof(p->s.fp) ) {
			c->state = CAP_STATE_STOP;
			break;
		}

		if ( svr && !feof(p->s.fp) ) {
			snd = &p->s;
			rcv = &p->c;
		}else{
			snd = &p->c;
			rcv = &p->s;
		}

		do_packet(p, snd, rcv, len);
	}
}

struct capdev firestorm_capdevs[]={
	{
		.name = "fagrouter",
		.desc = "TCP stream reassembly test rig",
		.init = fag_init,
		.end = fag_end,
		.go = fag_go,
		.live = 1,
	},
	{.name = NULL},
};

struct plugin_desc firestorm_plugin_desc = {
	.size		= sizeof(firestorm_plugin_desc),
	.name		= "capture.fagrouter",
	.desc		= "TCP stream test rig",
	.author_name	= "Gianni Tedesco",
	.author_email	= "gianni@scaramanga.co.uk",
	.ver_major	= 0,
	.ver_minor	= 1,
};
