/*
 * This file is part of Firestorm NIDS
 * Copyright (c) 2002,2003,2004 Gianni Tedesco <gianni@scaramanga.co.uk>
 * This program is released under the terms of the GNU GPL version 2
 *
 * http - Firestorm HTTP decode plugin
 *
 * TODO:
 *  o Add data layers
 *  o Put CONNECT support back in
 *  o Support chunked transfer encoding
 *  o Need to test pipelining properly
*/
#include <firestorm.h>
#include <dotscara/flbuf.h>
#include <dotscara/fdctl.h>

#include <limits.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <arpa/inet.h>
#include <fcntl.h>
#include <unistd.h>

#include <args.h>
#include <plugin.h>
#include <alert.h>
#include <signature.h>
#include <decode.h>
#include <preproc.h>
#include <strtouint.h>
#include <idx.h>
#include <matcher.h>
#include <field.h>

#include "http_normalize.h"

static struct generator http_gen = init_generator("http", &http_tb);
static struct alert a_invalid_port = {
	.alert = "Invalid Port",
	.sid = 1, .rev = 0,
	.priority = 5,
	.gen = &http_gen,
	.order = ALERT_BUILTIN,
};
static unsigned int nads;
extern struct proto http_p;

/* Accepted HTTP versions, must correspond with the  */
static struct {
	int maj, min;
}http_proto_vers[HTTP_VER_MAX]={
	[HTTP_VER_UNKNOWN] {-1, -1},
	[HTTP_VER_0_9] {0, 9},
	[HTTP_VER_1_0] {1, 0},
	[HTTP_VER_1_1] {1, 1},
};

/* HTTP Decode Control Block */
struct http_dcb {
	const char *label;
	void (*fn)(struct http_dcb *, struct fvec *);
	union {
		struct fvec *vec;
		int *val;
	}u;
};

/* Get a version code for a given major and minor version */
static inline unsigned int http_vers_idx(int maj, int min)
{
	int i;

	for(i=1; i < HTTP_VER_MAX; i++) {
		if ( http_proto_vers[i].maj == maj &&
			http_proto_vers[i].min == min )
		return i;
	}

	return HTTP_VER_UNKNOWN;
}

/* Parse and HTTP version string (eg: "HTTP/1.0") */
static int http_proto_version(struct fvec *str)
{
	const char *s = str->ptr;
	int maj, min;
	int ret;

	if ( str->ptr == NULL )
		return HTTP_VER_UNKNOWN;

	if ( str->len != 8 ) /* sizeof("HTTP/X.Y") */
		return HTTP_VER_UNKNOWN;

	if ( memcmp(s, "HTTP/", 5) )
		return HTTP_VER_UNKNOWN;

	if ( s[6] != '.' )
		return HTTP_VER_UNKNOWN;

	if ( !isdigit(s[5]) || !isdigit(s[7]) )
		return HTTP_VER_UNKNOWN;

	maj = s[5] - '0';
	min = s[7] - '0';
	ret = http_vers_idx(maj, min);

	return ret;
}

static void htype_string(struct http_dcb *h, struct fvec *v)
{
	if ( !h->u.vec )
		return;

	h->u.vec->ptr = v->ptr;
	h->u.vec->len = v->len;
}

static void htype_present(struct http_dcb *h, struct fvec *v)
{
	*h->u.val = 1;
}

static void htype_int(struct http_dcb *h, struct fvec *v)
{
	unsigned int val;
	size_t len;

	len = vtouint(v, &val);
	if ( !len || len > INT_MAX ) {
		*h->u.val = val;
	}else{
		*h->u.val = -1;
	}
}

/* Same as int but ensure input is only 3 digits */
static void htype_code(struct http_dcb *h, struct fvec *v)
{
	if ( v->len != 3 ) {
		*h->u.val = -1;
		return;
	}

	htype_int(h, v);
}

/* Constructor for HTTP state */
static void http_ctor(void *priv)
{
	struct http_flow *f=priv;
	f->client.state = HTTPS_HEADER;
	f->server.state = HTTPS_HEADER;
}

/* Hex decode a single ASCII char */
static inline int hexchar(unsigned int c)
{
	if ( c - '0' < 10 )
		return c - '0';
	if ( c - 'a' < 6 )
		return (c - 'a') + 10;
	if ( c - 'A' < 6 )
		return (c - 'A') + 10;

	return -1;
}

static int http_dprint(struct packet *p, int l, char *buf, size_t buflen)
{
	struct http_request *h;
	char *obuf=buf;

	h = p->layer[l].session;

	if ( h == NULL )
		return snprintf(buf, buflen, "[No session info]");

	/* 27 = 13+12+2 = bits of strings */
	if ( h->type == HTTP_REQUEST ) {
		if ( buflen < 27 + h->uri.len + h->method.len )
			return snprintf(buf, buflen, "[Buffer too small]");

		memcpy(buf, "http_method='", 13);
		buf+=13;

		memcpy(buf, h->method.ptr, h->method.len);
		buf+=h->method.len;

		memcpy(buf, "' http_uri='", 12);
		buf+=12;

		memcpy(buf, h->uri.ptr, h->uri.len);
		buf+=h->uri.len;

		memcpy(buf, "'", 2); /* including trailing null */
	}else{
		struct http_response *r = p->layer[l].session;
		snprintf(buf, buflen, "http_code='%i'", r->code);
	}

	return (buf-obuf);
}

static char *nads_emu = NULL;
static int cb_emu(struct arg *a, void *priv)
{
	if ( nads_emu ) {
		mesg(M_ERR, "http_normalize: %s emulator already specified",
			nads_emu);
		return 0;
	}

	nads_emu = strdup(a->val.v_str);

	if ( nads_emu == NULL ) {
		mesg(M_ERR, "http_normalize: strdup(): %s", sys_err());
		return 0;
	}

	return 1;
}

static int cbfn(const char *fn, u_int8_t *buf, size_t line, size_t len)
{
	char *port, *emu;
	unsigned int prt;
	u_int32_t ip;

	while ( *buf && isspace(*buf) ) {
		buf++;
		len--;
	}

	if ( *buf == '\0' || *buf == '#' )
		return 1;

	port = strchr(buf, ':');
	if ( port == NULL )
		goto parse_err;
	*port++ = '\0';

	emu = strchr(port, ':');
	if ( emu == NULL )
		goto parse_err;
	*emu++ = '\0';

	if ( *emu == '\0' ) {
		mesg(M_ERR, "http_server_map: %s:%u: No emulator specified",
				fn, line);
		return 1;
	}

	if ( port ) {
		if ( strtouint(port, &prt) )
			goto parse_err;
		if ( prt & ~0xffff )
			goto parse_err;
	}else{
		prt = HTTP_DEFAULT_PORT;
	}

	if ( !strtoip(buf, &ip) )
		goto parse_err;

	if ( !nads_webserver(ip, prt, emu) ) {
		mesg(M_ERR, "http_server_map: %s:%u: %s", fn, line,
				sys_err2("Emulator not found"));
		return 0;
	}

	return 1;
parse_err:
	mesg(M_ERR, "http_server_map: %s:%u: Parse error", fn, line);
	return 0;
}

static int cbfn2(const char *fn, u_int8_t *buf, size_t line, size_t len)
{
	mesg(M_ERR, "http_server_map: %s:%u: Line too long", fn, line);
	return 0;
}

static int http_server_map(const char *fn)
{
	struct flbuf b;
	int ret = 0;
	int fd;

	fd = open(fn, O_RDONLY);
	if ( fd < 0 ) {
		mesg(M_ERR, "http_server_map: %s: open(): %s",
				fn, sys_err());
		return 0;
	}

	if ( !flbuf_attach(&b, fd, 4096, (flbuf_cbfn_t)cbfn,
				(flbuf_cbfn_t)cbfn2, (void *)fn) ) {
		mesg(M_ERR, "http_server_map: %s: flbuf_attach(): %s",
				fn, sys_err());
		goto err;
	}

	if ( !flbuf_read(&b) ) {
		mesg(M_ERR, "http_server_map: %s: flbuf_read(): %s",
				fn, sys_err());
		goto err;
	}

	ret = 1;
err:
	fd_close(fd);
	return ret;
}

static int cb_smap(struct arg *a, void *priv)
{
	return http_server_map(a->val.v_str);
}

static int nads_init(const char *args)
{
	int freeit = 0;
	struct arg nads_args[]={
		{"emulate", ARGTYPE_STRING, cb_emu},
		{"servermap", ARGTYPE_STRING, cb_smap},
		{NULL, ARGTYPE_NOP, NULL},
	};

	if ( nads ) {
		mesg(M_ERR, "http_normalize: Can't add http_normalize twice!");
		return 0;
	}

	if ( !http_normalize_init() )
		return 0;

	if ( args ) {
		switch ( args_parse(nads_args, args, NULL) ) {
		case -1:
			mesg(M_ERR, "http_normalize: Parse Error: %s", args);
		case 0:
			return 0;
		default:
			break;
		}
	}

	if ( nads_emu ) {
		freeit = 1;
	}else{
		nads_emu = "IIS";
	}

	if ( !nads_default_webserver(nads_emu) ) {
		mesg(M_ERR, "http_normalize: Unable to emulate '%s' server",
			nads_emu);
		mesg(M_INFO, "http_normalize: Only 'IIS' and 'Apache' "
			"emulations are available currently");
		return 0;
	}

	mesg(M_INFO, "http_normalize: default emulator: %s", nads_emu);

	if ( freeit )
		free(nads_emu);

	nads = 1;
	return 1;
}

/* Check if this header is one we want to store */
static inline void dispatch_hdr(struct http_dcb *d,
				struct fvec *k,
				struct fvec *v)
{
	for(; d->label; d++) {
		if ( vstrcmp(k, d->label) )
			continue;
		d->fn(d, v);
	}
}

/* Actually parse an HTTP request */
static size_t http_decode_buf(struct http_dcb *d, char *p, char *end)
{
	char *cur;
	struct fvec hv[3]; /* method, url, proto */
	struct fvec k,v;
	int i=0;
	int state=0;
	int ret=0;

	hv[0].ptr=NULL;
	hv[1].ptr=NULL;
	hv[2].ptr=NULL;

	for(cur=p; cur<end; cur++) {
		switch ( state ) {
		case 0:
			if ( *cur != ' ' ) {
				state=1;
				hv[i].ptr=cur;
				hv[i].len=0;
			}
			break;
		case 1:
			hv[i].len++;
			switch(*cur) {
			case ' ':
				if ( i<2 ) {
					state=0;
					i++;
				}
				break;
			case '\n':
				if ( hv[i].len && *(cur-1)=='\r' )
					hv[i].len--;
				k.ptr = cur + 1;
				k.len = 0;
				state = 2;
				break;
			}
			break;
		case 2:
			if ( *cur == ':' ) {
				state = 3;
				break;
			}else if ( *cur=='\n' ) {
				ret = (cur - p) + 1;
				cur = end;
			}
			k.len++;
			break;
		case 3:
			if ( *cur != ' ' ) {
				v.ptr = cur;
				v.len = 0;
				state = 4;
			}
			break;
		case 4:
			v.len++;
			if ( *cur == '\n' ) {
				if ( v.len && *(cur-1)=='\r' ) {
					v.len--;
				}
				dispatch_hdr(d+3, &k, &v);
				k.ptr = cur + 1;
				k.len = 0;
				state = 2;
				break;
			}
			break;
		}
	}

	if ( hv[0].ptr==NULL || hv[1].ptr==NULL )
		return 0;

	/* Setup method/url/proto */
	d[0].fn(&d[0], &hv[0]);
	d[1].fn(&d[1], &hv[1]);
	d[2].fn(&d[2], &hv[2]);

	return ret;
}

/* Return 0 if we need to start reassembling */
static size_t http_sdecode(struct packet *p, char *buf, size_t len)
{
	struct tcp_session *s;
	struct http_fside f;
	char *begin, *end;
	int req=1;

	begin = buf;
	end = buf + len;

	/* Get the state information */
	s = p->layer[p->llen-1].session;
	if ( s->flow ) {
		struct http_flow *flow = s->flow;
		if ( p->layer[p->llen-1].flags & FLAG_TCP_2SVR ) {
			f.state = flow->client.state;
			f.len = flow->client.len;
		}else{
			f.state = flow->server.state;
			f.len = flow->server.len;
			req = 0;
		}
	}else{
		f.state = HTTPS_HEADER;
		if ( p->layer[p->llen-1].flags & FLAG_TCP_2SVR)
			req = 0;
	}

	/* Parse as much data as possible */
again:
	if ( f.state == HTTPS_HEADER ) {
		size_t hlen;
		int clen = -1;
		struct http_dcb dcb[]={
			{"method", htype_string, {.vec=NULL}},
			{"uri", htype_string, {.vec=NULL}},
			{"protocol", htype_string, {.vec=NULL}},
			{"Content-Length", htype_int, {.val=&clen}},
			{NULL,}
		};

		hlen = http_decode_buf(dcb, begin, end);
		if ( !hlen )
			return begin - buf;

		begin += hlen;

		if ( clen > 0 ) {
			f.state = HTTPS_CONTENT;
			f.len = clen;
		}else if ( !req && clen<0 ){
			f.state = HTTPS_CLOSING;
			return len;
		}
	}else if ( f.state == HTTPS_CONTENT ) {
		size_t rlen = end-begin;

		if ( rlen >= f.len ) {
			begin += f.len;
			f.len = 0;
			f.state = HTTPS_HEADER;
		}else{
			f.len -= rlen;
			begin += rlen;
		}
	}else if ( f.state == HTTPS_CLOSING ) {
		return len;
	}else{
		mesg(M_DEBUG, "Aaaargh");
		return len;
	}

	if ( begin < end )
		goto again;

	if ( begin > end )
		begin = end;

	return begin - buf;
}

/* Parse and pump an HTTP request */
static size_t http_request(struct packet *p, struct http_fside *f,
			char *begin, char *end)
{
	int clen=-1;
	size_t hlen;
	struct http_request r;
	struct fvec pv = {.ptr = NULL};
	int prox = 0;
	size_t i;
	int state, do_host;
	struct http_dcb dcb[]={
		{"method", htype_string, {.vec = &r.method}},
		{"uri", htype_string, {.vec = &r.uri}},
		{"protocol", htype_string, {.vec = &pv}},
		{"host", htype_string , {.vec = &r.host}},
		{"Content-Length", htype_int, {.val = &clen}},
		{"Proxy-Connection", htype_present, {.val = &prox}},
		{NULL,}
	};

	memset(&r, 0, sizeof(r));

	/* Do the decode */
	hlen = http_decode_buf(dcb, begin, end);
	if ( !hlen )
		return 0;

	/* Fill in other fields */
	r.type = HTTP_REQUEST;
	r.port = HTTP_DEFAULT_PORT;
	r.proto_vers = http_proto_version(&pv);

	/* Attach the session data now we have validated */
	p->layer[p->llen].session = &r;
	p->llen++;

	/* Update flow state */
	if ( clen > 0 ) {
		f->state = HTTPS_CONTENT;
		f->len = clen;
	}

	/* TODO: Implement */
	if ( !vstrcmp(&r.method, "CONNECT") ) {
	}

	/* Strip out Request-URI to just abs_path, Filling in host
	 * information if there was no host header
	 */
	if ( !prox ) {
		if ( r.uri.len < 7 || memcmp(r.uri.ptr, "http://", 7) )
			goto done;
	}

	for(i=state=do_host=0; i < r.uri.len; i++) {
		if ( state == 0 ) {
			if ( ((char *)r.uri.ptr)[i] == ':' )
				state++;
		}else if ( state >= 1 && state < 4 ) {
			if ( ((char *)r.uri.ptr)[i] == '/' ) {
				if ( state == 3 && !r.host.ptr ) {
					r.host.ptr = r.uri.ptr + i + 1;
					r.host.len = 0;
					do_host = 1;
				}
				state++;
			}else if ( do_host ) {
				r.host.len++;
			}
		}else{
			i -= 1;
			break;
		}
	}

	if ( state < 3 )
		goto done;

	r.uri.ptr += i;
	r.uri.len -= i;

	if ( r.uri.len == 0 )
		r.uri.ptr = NULL;

done:
	/* Extract the port from the host header */
	if ( r.host.ptr ) {
		size_t i;
		struct fvec port = { .ptr = NULL };
		unsigned int prt;

		for(i=r.host.len; i; i--) {
			if (  ((u_int8_t *)r.host.ptr)[i-1] == ':' ) {
				port.len = r.host.len - i;
				port.ptr = r.host.ptr + i;
				r.host.len = i - 1;
				break;
			}
		}

		if ( port.len ) {
			if ( vtouint(&port, &prt) == port.len ) {
				if ( prt & ~0xffffUL ) {
					alert_tag(p, &a_invalid_port, -1);
				}else{
					r.port = prt;
				}
			}
		}
	}

	/* rfc2616: An empty abs_path is equivalent to an abs_path of "/" */
	if ( !r.uri.ptr ) {
		r.uri.ptr = "/";
		r.uri.len = 1;
	}

	if ( nads )
		http_normalize(p, &r);

	dispatch(p);
	return hlen;
}

/* Parse and pump an HTTP response */
static size_t http_response(struct packet *p, struct http_fside *f,
			char *begin, char *end)
{
	int clen=-1;
	size_t hlen;
	struct http_response r;
	struct fvec pv = {.ptr = NULL};
	struct http_dcb dcb[]={
		{"protocol", htype_string, {.vec = &pv}},
		{"code", htype_code, {.val = &r.code}},
		{"msg", htype_string, {.vec = NULL}},
		{"Content-Length", htype_int, {.val = &clen}},
		{"Server", htype_string, {.vec = &r.server}},
		{NULL,}
	};

	memset(&r, 0, sizeof(r));

	hlen = http_decode_buf(dcb, begin, end);
	if ( !hlen )
		return 0;

	p->layer[p->llen].session = &r;
	p->llen++;

	r.type = HTTP_RESPONSE;
	r.proto_vers = http_proto_version(&pv);

	if ( clen > 0 ) {
		f->state = HTTPS_CONTENT;
		f->len = clen;
	}else if ( clen < 0 ) {
		f->state = HTTPS_CLOSING;
	}

	dispatch(p);
	return hlen;
}

/* HTTP decode callback */
static void http_decode(struct packet *p)
{
	struct http_flow noflow;
	struct http_flow *flow;
	struct tcp_session *s;
	struct http_fside *f;
	char *begin, *end;
	int olen = p->llen;
	int dispatches=0;
	size_t (*fn)(struct packet *, struct http_fside *, char *, char *);

	/* Grab the data portion */
	begin = p->layer[p->llen].h.raw;
	end = p->layer[p->llen-2].h.raw +
		ntohs(p->layer[p->llen-2].h.ip->tot_len);

	if ( begin >= end )
		goto done;

	/* Setup state information */
	s = p->layer[p->llen-1].session;
	if ( s->flow ) {
		flow = s->flow;
	}else{
		flow = &noflow;
		http_ctor(flow);
	}

	/* Get the right side */
	if ( p->layer[p->llen-1].flags & FLAG_TCP_2SVR ) {
		f = &flow->client;
		fn = http_request;
	}else{
		f = &flow->server;
		fn = http_response;
	}

again:
	if ( f->state == HTTPS_HEADER ) {
		size_t hlen;

		p->llen = olen;
		hlen = fn(p, f, begin, end);
		if ( !hlen )
			goto done;

		dispatches++;

		begin += hlen;
	}else if ( f->state == HTTPS_CONTENT ) {
		size_t rlen = end-begin;

		if ( rlen >= f->len ) {
			begin += f->len;
			f->len = 0;
			f->state = HTTPS_HEADER;
		}else{
			f->len -= rlen;
			begin += rlen;
		}
	}else if ( f->state == 3) {
		goto done;
	}else{
		mesg(M_DEBUG, "corrupt flow");
	}

	if ( begin < end )
		goto again;

done:
	if ( !dispatches )
		dispatch(p);
}

static void http_session_free(struct http_request *r)
{
	if ( (r->type == HTTP_REQUEST) && (r->free != NULL) )
		free(r->free);
	free(r);
}

static int get_method(struct packet *p, int l, struct fvec *vec)
{
	struct http_request *s;

	l += 2;
	if ( l >= p->llen )
		return 0;
	if ( p->layer[l].proto != &http_p )
		return 0;
	if ( p->layer[l].session == NULL )
		return 0;

	s = p->layer[l].session;

	if ( s->type != HTTP_REQUEST )
		return 0;

	vec->ptr = s->method.ptr;
	vec->len = s->method.len;

	return 1;
}

struct field http_fields[] = {
	{
		.name = "http.method",
		.desc = "HTTP Method",
		.type = IDX_SKUNK_NOCASE,
		.get = {.str = get_method},
		.cost = MCOST_APP,
		.ofs = 1,
	},
	{.name = NULL}
};

static struct proto http_p={
	.name = "http",
	.sdecode = http_sdecode,
	.decode = http_decode,
	.print = http_dprint,
	.flow_len = sizeof(struct http_flow),
	.flow_ctor = http_ctor,
	.serialize = http_serialize,
	.deserialize = http_deserialize,
	.free = (proc_session_free)http_session_free,
	.fields = http_fields,
};

static struct proto_req http_r[]={
	proto_request("tcp", __constant_htons(80)),
	proto_request("tcp", __constant_htons(3128)),
	proto_request("tcp", __constant_htons(8080)),
	null_request()
};

struct proto_table firestorm_protocols[] = {
	{.proto = &http_p, .preq = http_r},
	{.proto = NULL}
};

struct preproc firestorm_preprocs[] = {
	{.name = "http_normalize", .init = nads_init},
	{.name = NULL },
};

struct plugin_desc firestorm_plugin_desc = {
	.size		= sizeof(firestorm_plugin_desc),
	.name		= "decode.http",
	.desc		= "Hyper-text Transfer Protocol",
	.author_name	= "Gianni Tedesco",
	.author_email	= "gianni@scaramanga.co.uk",
	.ver_major	= 0,
	.ver_minor	= 2,
	.fini		= http_normalize_fini,
};
