/*
 * macwatch - Firestorm arp-shannanigans detector
 * Copyright (c) 2003-2004 John Leach <john@johnleach.co.uk>
 * Released under the terms of the GNU GPL v2
 * $Id: macwatch.c 455 2004-04-23 15:08:07Z scara $
 * 
 * TODO:
 * o consider order of alerts (only one alert per packet is allowed)
 * o DoS protection (hi/lo memory watermarks with LFU lists) (should we actually remove older entries?)
 * o Saving of tuple database in between restarts
 * o ARP sweep detection?
 * o MAC flooding detection
 * o duplicate detection alert expiry time
 * o Alert on replies with no requests
 * o handle requests from 0.0.0.0 (dhcp/bootp (RFC 1541) suggests verifying lease using arp request FROM 0.0.0.0)?
 * o IP handler
 * o Router tracker, track MAC addresses of non local IP traffic?
 * o alert on incorrect address lengths
 * o Other protocols?  Netbios names?
 * o Only one entry per protocol address?
 * o Keep track of requests+replies.  Ensure from different stations.  Alert on unsolicited 	
 * 
 * DONE:
 * o Better IP hash
 * o Variable sized hardware/protocol address support
 * o Check destination addresses on replies
 * o IPX, 8 byte protocol address 
 * o Improved hash function (upgraded to 1a)
 * o hln/pln matching bug (could match on a portion of a larger h/paddr)
 * o verify ethernet address against arp hardware address
 * o check arp decoder layer flags to make sure addresses are available
 * o Gratuitous arp detection (sender/target protocol addresses match)
 * o Unicast request detection (toggleable to disable as routers often use this as cache verification)
 * o Code tidy up
 * o Toggleable new address reporting
 * o Replace is_etherbcast and pkttypeflag macros with inline functions
 * o check malloc return codes
*/

#include <firestorm.h>

#include <sys/socket.h>
#include <netinet/in.h>
#include <arpa/inet.h>

#include <plugin.h>
#include <args.h>
#include <alert.h>
#include <signature.h>
#include <decode.h>
#include <preproc.h>

/* Configuration */
static unsigned int check_gratuitous;
static unsigned int check_unirequest;
static unsigned int alertnew;

static struct arg args[] = {
      {"check_gratuitous", ARGTYPE_PBOOL, NULL, {vp_bool:&check_gratuitous}},
      {"check_unirequest", ARGTYPE_PBOOL, NULL, {vp_bool:&check_unirequest}},
      {"alertnew", ARGTYPE_PBOOL, NULL, {vp_bool:&alertnew}},
      {NULL, ARGTYPE_NOP, NULL}
};

/* Stats */
static unsigned int newaddrs, duplicates;

/* Protocol and hardware address sizes */
#define MAX_PLN 	16	/* to accommodate up to ipv6 (128bit) */
#define MAX_HLN 	6

static u_int8_t etherbcast[6] = { 0xff, 0xff, 0xff, 0xff, 0xff, 0xff };

/* The macwatch address struct */
struct arpaddr {
	u_int8_t haddr[MAX_HLN];	/* Hardware address */
	u_int8_t paddr[MAX_PLN];	/* Protocol address */
	u_int8_t hln, pln;	/* Hardware/protocol address length */
	u_int16_t hrd, proto;	/* ARP protocol type */
	/* FIXME?  Variable size addresses could be faster due to "improved 
	 * locality" cache efficiency */
	struct arpaddr *next;
};

#define ARPLENGTH sizeof(struct pkt_arphdr)
#define MW_ADDR_HASH_SIZE 256
#define MW_ADDR_HASH_MASK (MW_ADDR_HASH_SIZE-1)

static struct arpaddr *addrhash[MW_ADDR_HASH_SIZE];
static proc_decode ethdecode;

/* Check that specified flag is set and assured */
static inline int pkttypeflag(u_int32_t flags, u_int32_t flag)
{
       return ( ( flags & ( flag | FP_PKTTYPE ) ) == ( flag | FP_PKTTYPE ) );
}

/* Check if address is ethernet broadcast (6 X 0xff) */
static inline int is_etherbcast(u_int8_t *addr)
{
       return ( addr && memcmp( addr, etherbcast, sizeof( etherbcast ) ) == 0 );
}


/* Alert stuff */
static struct tokenbucket mw_tb;
static struct generator mw_gen = init_generator("macwatch", &mw_tb);
static struct alert alert_newpa = {
	.alert = "New protocol address",
	.sid = 0, .rev = 0,
	/* This *should* be a lower priority but Gianni hasn't
	   decided on a priority policy yet */
	.priority = 5,
	.gen = &mw_gen,
	.order = ALERT_BUILTIN,
};
static struct alert alert_dupha = {
	.alert = "Duplicate hardware address",
	.sid = 1, .rev = 0,
	.priority = 5,
	.gen = &mw_gen,
	.order = ALERT_BUILTIN,
};
static struct alert alert_srcnomatch = {
	.alert = "Hardware src doesn't match ethernet frame",
	.sid = 2, .rev = 0,
	.priority = 5,
	.gen = &mw_gen,
	.order = ALERT_BUILTIN,
};
static struct alert alert_dstnomatch = {
	.alert = "Hardware dst doesn't match ethernet frame",
	.sid = 3, .rev = 0,
	.priority = 5,
	.gen = &mw_gen,
	.order = ALERT_BUILTIN,
};
static struct alert alert_bcastreply = {
	.alert = "Broadcast reply",
	.sid = 4, .rev = 0,
	.priority = 5,
	.gen = &mw_gen,
	.order = ALERT_BUILTIN,
};
static struct alert alert_gratuitous = {
	.alert = "Gratuitous arp",
	.sid = 5, .rev = 0,
	.priority = 5,
	.gen = &mw_gen,
	.order = ALERT_BUILTIN,
};
static struct alert alert_unirequest = {
	.alert = "Unicast request",
	.sid = 6, .rev = 0,
	.priority = 5,
	.gen = &mw_gen,
	.order = ALERT_BUILTIN,
};

/* Fowler/Noll/Vo 1a hash taken from SkunkDB */
static inline u_int32_t skunk_hash(u_int8_t *data, size_t len)
{
	u_int32_t h = 2166136261UL;
	while (len) {
		h = (h * 16777619) ^ *data++;
		--len;
	}
	return h;
}

/* Add tuple to database, alert accordingly */
static void add_addr(void *pa, void *ha, struct packet *pkt, u_int8_t hln,
		     u_int8_t pln, u_int16_t hrd, u_int16_t proto)
{
	struct arpaddr *mw;
	struct arpaddr *mwn;
	u_int32_t tuple_count = 0;	/* Number of duplicate tuples */
	u_int32_t hashval;

	hashval = skunk_hash(pa, pln) & MW_ADDR_HASH_MASK;

	for (mw = addrhash[hashval]; mw; mw = mw->next) {
		if ((mw->proto == proto) && (mw->hrd == hrd) &&
		    (mw->pln == pln) && (mw->hln == hln) &&
		    (memcmp(pa, &mw->paddr, pln) == 0)) {
			/* Seen already, stop searching, don't add */
			if (memcmp(ha, &mw->haddr, hln) == 0)
				return;
			/* MAC address differs? count it */
			else
				tuple_count++;
		}
	}

	/* Reached end of linked list, create new entry */
	mwn = malloc(sizeof(struct arpaddr));
	if (!mwn) {
		mesg(M_ERR,"macwatch: malloc(): %s", sys_err());
		return;
	}
	memcpy(&mwn->haddr, ha, hln);
	memcpy(&mwn->paddr, pa, pln);
	mwn->hln = hln;
	mwn->pln = pln;
	mwn->hrd = hrd;
	mwn->proto = proto;
	mwn->next = addrhash[hashval];
	addrhash[hashval] = mwn;

	/* Alert accordingly */
	if (tuple_count) {
		alert_tag(pkt, &alert_dupha, -1);
		duplicates++;
	} else {
		if (alertnew) alert_tag(pkt, &alert_newpa, -1);
		newaddrs++;
	}
}

/* Pre-process arp packets */
static void macwatch_arpwatcher(struct packet *pkt, int l, void *priv)
{
	u_int8_t hln, pln;	/* address lengths */
	u_int16_t hrd, proto;
	u_int8_t *hw_src = NULL, *hw_dst = NULL;	/* hardware layer addresses */
	u_int8_t *arph_src = NULL, *arph_dst = NULL;	/* arp hardware addresses */
	u_int8_t *arpp_src = NULL, *arpp_dst = NULL;	/* arp protocol addresses */
	void *ofs = pkt->layer[l].h.raw;

	hln = pkt->layer[l].h.arp->hln;
	pln = pkt->layer[l].h.arp->pln;
	hrd = pkt->layer[l].h.arp->hrd;

	/* host order conversion unnecessary */
	proto = pkt->layer[l].h.arp->proto;

	/* Check hw and proto address sizes  */
	/* XXX: Alert or log here too?. */
	if ((hln > MAX_HLN) || (pln > MAX_PLN))
		return;

	/* Sender address data available? */
	if (pkt->layer[l].flags & FLAG_ARP_SENDER) {
		arph_src = ofs + ARPLENGTH;
		arpp_src = ofs + ARPLENGTH + hln;
	}

	/* Target address data available? */
	if (pkt->layer[l].flags & FLAG_ARP_TARGET) {
		arph_dst = ofs + ARPLENGTH + hln + pln;
		arpp_dst = ofs + ARPLENGTH + hln + pln + hln;
	}

	/* Ethernet specific checks */
	if (l - 1 >= 0 && pkt->layer[l - 1].proto->decode == ethdecode) {
		hw_src = pkt->layer[l - 1].h.eth->src;
		hw_dst = pkt->layer[l - 1].h.eth->dst;

		/* Non matching hardware src */
		if (arph_src && (memcmp(hw_src, arph_src, hln) != 0)) {
			alert_tag(pkt, &alert_srcnomatch, -1);
		}
	}

	/* Gratuitous arp */
	if (check_gratuitous && arpp_src && arpp_dst &&
	    (memcmp(arpp_src, arpp_dst, hln) == 0)) {
		alert_tag(pkt, &alert_gratuitous, -1);
	}

	switch (ntohs(pkt->layer[l].h.arp->op)) {
	case 1:
		/* ARP Request */
		if (arph_src)
			add_addr(arpp_src, arph_src, pkt, hln, pln, hrd, proto);

		if (check_unirequest) {
			/* Unicast request */
			if (pkttypeflag(pkt->flags, FP_HOST))
				alert_tag(pkt, &alert_unirequest, -1);
			else if (hw_dst && !is_etherbcast(hw_dst))
				alert_tag(pkt, &alert_unirequest, -1);
		}
		break;
	case 2:
		/* ARP Reply */
		if (arph_src)
			add_addr(arpp_src, arph_src, pkt, hln, pln, hrd, proto);
		if (arph_dst)
			add_addr(arpp_dst, arph_dst, pkt, hln, pln, hrd, proto);

		/* Broadcast reply */
		if (pkttypeflag(pkt->flags, FP_BROADCAST))
			alert_tag(pkt, &alert_bcastreply, -1);

		/* Ethernet specific checks */
		if (hw_dst) {
			/* Broadcast reply */
			if (is_etherbcast(hw_dst))
				alert_tag(pkt, &alert_bcastreply, -1);
			/* Hardware addresses differ */
			if (arph_dst
			    && !(memcmp(hw_dst, arph_dst, hln) == 0))
				alert_tag(pkt, &alert_dstnomatch, -1);
		}
		break;
	}
}

static void macwatch_free(void *priv)
{
	int i;
	unsigned int ips=0, hashes=0;
	struct arpaddr *mw, *mwp;

	for (i = 0; i < MW_ADDR_HASH_SIZE; i++)
		if (addrhash[i] != NULL) {
			hashes++;
			for (mw = addrhash[i]; mw != NULL;) {
				ips++;
				mwp = mw;
				mw = mw->next;
				free(mwp);
			}
		}
	mesg(M_INFO, "macwatch: %i new addresses, %i duplicates", newaddrs, duplicates);
	mesg(M_INFO, "macwatch: %i address in %i hashes", ips, hashes);
	
	return;
};

static int macwatch_init(const char *str)
{
	struct proto *p;

	if (str && args_parse(args, str, NULL) <= 0) {
		mesg(M_ERR, "macwatch: parse error: %s", str);
		return 0;
	}

	if ((p = decode_proto("ethernet"))) {
		ethdecode = p->decode;
	}

	return preproc_activate("arp", macwatch_arpwatcher,
				macwatch_free, NULL);
}

struct preproc firestorm_preprocs[] = {
	{.name = "macwatch", .init = macwatch_init},
	{.name = NULL}
};

struct plugin_desc firestorm_plugin_desc = {
	.size = sizeof(firestorm_plugin_desc),
	.name = "preproc.macwatch",
	.desc = "MAC address change watcher",
	.author_name = "John Leach",
	.author_email = "john@johnleach.co.uk",
	.ver_major = 0,
	.ver_minor = 6,
};
