/*
 * This file is part of Firestorm NIDS
 * Copyright (c) 2003,2004 Gianni Tedesco
 * Released under the terms of the GNU GPL version 2
 *
 * This file represents the index creation API.
*/

#include <firestorm.h>
#include <errno.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <unistd.h>
#include <fcntl.h>

#include <dotscara/fdctl.h>
#include <dotscara/fobuf.h>
#include <cat.h>
#include <elog.h>
#include <alert.h>
#include <signature.h>
#include <decode.h>
#include <target.h>
#include <elog_read.h>
#include <idx.h>
#include <i_idx.h>
#include <q_idx.h>
#include <q_filter.h>
#include <field.h>

static struct target *t;
static void *t_priv;
static const char idx_suffix[] = ".idx";

int cat_prepare(const char *fmt, const char *opts)
{
	/* Get the right output plugin */
	if ( !(t=target_find(fmt)) ) {
		mesg(M_ERR, "%s: Cannot find specified output format", fmt);
		return 0;
	}

	/* Initialise it with user specified arguments*/
	if ( !t->validate(opts, &t_priv) ) {
		mesg(M_ERR, "%s: failed to initialize output plugin", fmt);
		return 0;
	}

	return 1;
}

int cat_close(void)
{
	int ret;

	if ( !t )
		return 1;
	
	ret = t->cleanup(t_priv);

	t = NULL;
	t_priv = NULL;

	return ret;
}

static int cbfn_target(struct elog_pkt *p, void *data)
{
	return elog_pkt_alert(p, t->target, data);
}

static struct q_db *open_index(int fd, const char *fn)
{
	size_t fnlen = strlen(fn);
	char fnbuf[fnlen + sizeof(idx_suffix)];
	struct stat st1, st2;
	struct q_db *ret;
	int ifd;

	if ( fstat(fd, &st1) < 0 ) {
		mesg(M_ERR, "%s: fsync(): %s", fn, sys_err());
		return NULL;
	}

	snprintf(fnbuf, sizeof(fnbuf), "%s%s", fn, idx_suffix);
	ifd = open(fnbuf, O_RDONLY);
	if ( ifd < 0 ) {
		if ( errno == ENOENT ) {
			mesg(M_ERR, "First create an index file using "
				"firecat -i '%s'", fn);
		}else{
			mesg(M_ERR, "%s: open(): %s", fnbuf, sys_err());
		}
		return NULL;
	}

	if ( fstat(fd, &st2) < 0 ) {
		mesg(M_ERR, "%s: fsync(): %s", fnbuf, sys_err());
		__fd_close(fd);
		return NULL;
	}

	if ( st2.st_mtime < st1.st_mtime ) {
		mesg(M_ERR, "%s: %s is out of date", fn, fnbuf);
		__fd_close(fd);
		return NULL;
	}

	ret = q_idx_open(ifd);
	__fd_close(ifd);

	if ( ret == NULL )
		mesg(M_ERR, "%s: q_idx_open(): %s", fnbuf, sys_err2(NULL));

	return ret;
}

int cat_elog(int fd, const char *fn, void *query)
{
	struct q_filter *f = query;
	void *ep;
	int ret = 0;
	struct q_db *db = NULL;

	if ( !(ep=elog_new(fd)) ) {
		mesg(M_ERR, "%s: elog_new(): %s", fn, sys_err2(NULL));
		goto err;
	}

	if ( f ) {
		db = open_index(fd, fn);
		if ( db == NULL ) {
			goto err_free_elog;
		}
	}

	elog_set_alert_fn(ep, cbfn_target);

	if ( f ) {
		unsigned int i;

		if ( !q_filter_run(f, db) ) {
			mesg(M_ERR, "%s: q_filter_run(): %s",
					fn, sys_err2(NULL));
			goto err_free_db;
		}

		for(i=0; i < f->num_recs; i++) {
			if ( elog_pkt(ep, f->result[i], t_priv) <= 0 )
				goto err_free_result;
		}
	}else{
		if ( elog_run(ep, t_priv) <= 0 )
			goto err_free_db;
	}

	ret = 1;

err_free_result:
	if ( f ) {
		q_filter_free_result(f);
	}
err_free_db:
	if ( db )
		q_idx_close(db);
err_free_elog:
	elog_free(ep);
err:
	return ret;
}

int cat_prepare_query(const char *query, void **filter)
{
	struct q_filter *f;

	if ( query == NULL ) {
		*filter = NULL;
		return 1;
	}

	f = q_filter_new(0);
	if ( f == NULL ) {
		mesg(M_ERR, "q_filter_new: %s",
				sys_err2(NULL));
		return 0;
	}

	if ( !q_filter_query(f, query) ) {
		mesg(M_ERR, "q_filter_query: %s",
				sys_err2("Parse Error"));
		q_filter_free(f);
		return 0;
	}

	*filter = f;
	return 1;
}

void cat_free_query(void *qry)
{
	if ( qry == NULL )
		return;
	q_filter_free(qry);
}
