/*
* This file is part of Firestorm NIDS
* Copyright (c) 2002,2003,2004 Gianni Tedesco
* This program is released under the terms of the GNU GPL version 2
*
* This is the decode engine. We keep a big internal graph of all
* protocols and all their dependancies on each other. The actual
* decoding is done by plugins.
*/
#include <firestorm.h>
#include <dotscara/fobuf.h>
#include <args.h>
#include <alert.h>
#include <signature.h>
#include <decode.h>
#include <preproc.h>
#include <loader.h>
#include <idx.h>
#include <i_idx.h>
#include <q_idx.h>
#include <q_filter.h>
#include <field.h>
#include <string_hash.h>

/* libpcap DLT_* protocol */
static struct proto pcap_dlt_p = init_proto("__pcap_dlt", NULL, NULL);
const struct proto *pcap_dlt = &pcap_dlt_p;

/* Global protocol list */
#define DHASH_SIZE 127
static struct proto *protos[DHASH_SIZE];

/* Protocol request list */
static struct preq_list *preqs;

/* Find a protocol by name */
struct proto *decode_proto(const char *proto)
{
	struct proto *p;
	for(p=protos[string_hash(proto)%DHASH_SIZE]; p; p=p->next) {
		if ( !strcmp(p->name, proto) ) {
			return p;
		}
	}
	return NULL;
}

/* Capture plugins call this   */
struct proto *decode_subproto(const char *proto, unsigned int lt)
{
	struct proto *p;
	struct proto_child *c;

	if ( !(p=decode_proto(proto)) ) return NULL;

	for(c=p->children; c; c=c->next) {
		if ( c->id == lt ) {
			return c->proto;
		}
	}
	return NULL;
}

/* plugins call this to register protocols */
static int decode_add(struct proto *p, struct proto_req *r)
{
	struct preq_list *req;
	u_int32_t h;

	if ( p == NULL || r == NULL )
		return 0;
	
	if ( p->name == NULL || p->decode == NULL )
		return 0;

	if ( strlen(p->name) > PROTO_LEN ) {
		mesg(M_WARN, "decode: %s: protocol name is too long", p->name);
	}

	/* Check if we have already been added */
	if ( p->next || p->children || decode_proto((char *)p->name) ) {
		mesg(M_DEBUG,"decode: error: '%s' protocol is "
			"already registered", p->name);
		return 0;
	}

	/* XXX: Should we make all fields field_reg's instead? */
	if ( p->fields ) {
		struct field *ptr = p->fields;
		p->fields = NULL;
		field_add(p, ptr);
	}

	/* Save the requests for later, when all the protocols are done
	 * registering. We resolve them in decode_resolve_all()
	 */
	if ( !(req=calloc(1, sizeof(struct preq_list))) ) {
		firestorm_die_error("calloc");
	}

	req->proto = p;
	req->arr = r;
	req->next = preqs;
	preqs = req;

	/* Add to the hash table */
	h=string_hash((char *)p->name)%DHASH_SIZE;
	p->next=protos[h];
	protos[h]=p;

	return 1;
}

/* Resolve a protocol request */
static int decode_resolve(struct proto_req *req, struct proto *proto)
{
	struct proto_child *pc;
	struct proto *parent;

	if ( !(parent=decode_proto(req->name)) ) {
		return 0;
	}

	if ( !(pc=calloc(1, sizeof(struct proto_child))) ) {
		firestorm_die_error("calloc");
	}

	/* fill in the proto_child */
	pc->id=req->id;
	pc->proto=proto;

	/* append it to the rightful owner */
	pc->next=parent->children;
	parent->children=pc;

#if 0
	mesg(M_DEBUG,"resolve: %s:%u -> %s",
		parent->name,
		pc->id,
		pc->proto->name);
#endif

	return 1;
}

/* Build the graph of inter-dependencies
 * for all registered protocols */
static int decode_resolve_all(void)
{
	struct preq_list *foo, *bar;
	int i;

	for(foo=preqs; foo;) {
		bar=foo;
		foo=foo->next;

		for(i=0; bar->arr[i].name; i++) {
			decode_resolve(&bar->arr[i], bar->proto);
		}

		/* this is only a temporary structure
		 * after all */
		free(bar);
	}

	preqs=NULL;

	return 1;
}

static void decode_register(struct plugin *p, struct proto_table *t)
{
	for(; t->proto; t++)
		decode_add(t->proto, t->preq);
}

struct field *field_find(const char *name)
{
	struct field *f;
	struct proto *p;
	int i;

	for(f=field_builtins(); f; f=f->next) {
		if ( !strcmp(name, f->name) )
			return f;
	}

	for(i=0; i < DHASH_SIZE; i++) {
		for(p=protos[i]; p; p=p->next)
			for(f=p->fields; f; f=f->next)
				if ( !strcmp(name, f->name) )
					return f;
	}

	return NULL;
}

struct field *field_foreach(field_cbfn_t cbfn, void *priv)
{
	struct field *f;
	struct proto *p;
	int i;

	for(f=field_builtins(); f; f=f->next)
		if ( !(*cbfn)(f, priv) )
			return f;

	for(i=0; i < DHASH_SIZE; i++) {
		for(p=protos[i]; p; p=p->next)
			for(f=p->fields; f; f=f->next)
				if ( !(*cbfn)(f, priv) )
					return f;
	}

	return NULL;
}

void decode_load(void)
{
	loader_find_syms("firestorm_protocols", (loader_cb_t)decode_register);
	decode_resolve_all();
}

/* Destructor frees per-proto data structures */
static void decode_cleanup(struct packet *pkt, int l, void *priv)
{
	struct proto *p;
	int i;

	for(i=0; i < DHASH_SIZE; i++) {
		for(p=protos[i]; p; p=p->next)
			notifier_delete(&p->preproc);
	}
}

/* Constructor manually adds __pcap_dlt to the table */
void __private decode_init(void)
{
	u_int32_t h = string_hash((char *)pcap_dlt_p.name)%DHASH_SIZE;
	protos[h]=&pcap_dlt_p;
	notify_exit(decode_cleanup, NULL);
}
