/*
* This file is part of Firestorm NIDS
* Copyright (c) 2002,2003,2004 Gianni Tedesco
* This program is released under the terms of the GNU GPL version 2
*
* Alert Spool API
*  o spool_new() - create a new spool object
*  o spool_link() - link a spool in to the list
*  o spool_set_buf() - set the buffer, can only be called once
*  o spool_dump_header() - write file header
*  o spool_creat() - open and create logfile and write header
*  o spool_rotate() - rotate a logfile, creation of new file is optional
*  o spool_rotate_all() - rotate all log spools
*  o spool_rotate_check() - rotate all spools, only if necessary
*  o spool_check_old() - hack to rotate after a crash
*  o spool_delete() - close and destroy a spool object
*  o spool_hup() - respond to a HUP signal
*  o spool_init_all() - open all alert spools
*  o spool_cleanup() - flush and close all alert spools
*  o spool_build_packet() - construct an elog record
*  o spool_do_packet() - log a packet
*  o spool_next() - select the next spool
*  o spool_packet() - log a packet
*
* Stormwall Notification API
*  o stormwall_msg() - send a message to stormwall
*  o stormwall_open() - open the stormwall fifo
*  o stormwall_close() - close the stormwall fifo
*
* TODO
*  o Remove spool_build_packet and static buffers
*  o Resize serialization buffer if necessary
*  o Need to do clever things when writes fail
*  o Weighted round-robin distribution
*/
#include <firestorm.h>

#include <unistd.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <netinet/in.h>

#include <dotscara/fobuf.h>
#include <dotscara/fdctl.h>
#include <alert.h>
#include <signature.h>
#include <decode.h>
#include <stormwall.h>
#include <elog.h>
#include <elog_write.h>

/* Private API */
static void spool_packet(struct packet *p, int l, void *priv);
static void spool_cleanup(struct packet *pkt, int l, void *priv);

/* Stormwall API */
static int stormwall_msg(struct elog_spool *s, int msg);
static void stormwall_open(struct elog_spool *s);
static void stormwall_close(struct elog_spool *s);

/* A blank 4 byte area we can point to for zero padding */
static const char alert_padding[4];

/* List of all spools */
static struct elog_spool_head spools={
	(struct elog_spool *)&spools,
	(struct elog_spool *)&spools,
};

/* Current spool */
static struct elog_spool *cur_spool=(struct elog_spool *)&spools;

/* flush all alert spools */
static void spool_cleanup(struct packet *pkt, int l, void *priv)
{
	struct elog_spool *s=priv;

	*s->rotate_fn = '\0';
	mesg(M_INFO, "alert: %s: flushing to disk.", s->log_dir);

	spool_delete(s);
}

/* create a new spool object */
struct elog_spool *spool_new(void)
{
	struct elog_spool *ret;

	if ( !(ret=calloc(1, sizeof(*ret))) )
		return NULL;

	ret->stormwall=STORMWALL_NONE;
	ret->fifo_fd=-1;
	ret->max_bytes=SPOOL_DEFAULT_SIZE;
	ret->max_time=SPOOL_DEFAULT_MINS;

	return ret;
}

/* Check if rotation is necessary */
/* FIXME: year 2038 bug */
static inline int spool_must_rotate(struct elog_spool *s, struct timeval *tv)
{
	if ( s->size == 0 && s->tot_len > s->max_bytes )
		return 0;
	if ( s->max_bytes && (s->size + s->tot_len) > s->max_bytes )
		return 1;
	if ( s->max_time &&
		(unsigned int)tv->tv_sec > (tv->tv_sec + s->max_time) )
		return 1;
	return 0;
}

/* Link a spool in to the list */
void spool_link(struct elog_spool *s)
{
	s->next=spools.next;
	s->prev=(struct elog_spool *)&spools;
	spools.next->prev=s;
	spools.next=s;

	notify_exit(spool_cleanup, s);
}

/* Set the buffer size, can only be called once */
int spool_set_buf(struct elog_spool *s, size_t sz)
{
	fobuf_attach(&s->buf, -1, 0, sz);
	if ( sz==0 )
		s->nobuf = 1;
	return 1;
}

/* write file header */
static int spool_dump_header(struct elog_spool *s)
{
	struct efile_hdr fhdr;
	int hsize=sizeof(fhdr);

	/* Construct header */
	fhdr.magic=htonl(EF_MAGIC);
	fhdr.flags=0;
	fhdr.vers_minor=EF_VERS_MIN;
	fhdr.vers_major=EF_VERS_MAJ;

	/* Write header */
	if ( !fobuf_write(&s->buf, &fhdr, hsize) ) {
		mesg(M_CRIT,"elog: %s: fobuf_write(): %s",
			s->alert_fn, sys_err());
		return -1;
	}

	/* Setup the structures */
	s->size=sizeof(struct efile_hdr);

	return 0;
}

/* Open a new logfile and write the header */
static int spool_creat(struct elog_spool *s)
{
	int fd;

	/* Open up the log file */
	fd=open(s->alert_fn, O_WRONLY|O_CREAT|O_EXCL, 0600);
	fobuf_newfd(&s->buf, fd);

	if ( fd<0 ) {
		mesg(M_CRIT,"%s: open(): %s", s->alert_fn, sys_err());
		return -1;
	}

	return spool_dump_header(s);
}

/* All-in-one function to perform a log rotation */
static void spool_rotate(struct elog_spool *s, int new)
{
	struct stat st;

	/* Nothing to rotate */
	if ( spool_isempty(s) )
		return;

	fobuf_flush(&s->buf);

	/* Get the inode number before closing */
	if ( fstat(s->buf.fd, &st) ) {
		mesg(M_CRIT, "alert: fstat(): %s", sys_err());

		/* what are the chances of this happening twice
		 * in one second? Any better ideas? */
		st.st_ino=0xdeadbeef;
	}

	/* Flush and close the old one */
	fobuf_newfd(&s->buf, -1);

	/* Build a name to rotate the old file to */
	snprintf(s->rotate_fn, ALERT_FNLEN+1, "@%.8lx.%.16llx.elog",
		s->tv.tv_sec, (unsigned long long)st.st_ino);

	/* Rotate it */
	if ( rename(s->alert_fn, s->log_dir) ) {
		mesg(M_CRIT,"alert: rename(): %s", sys_err());
	}else{
		mesg(M_DEBUG,"logrotate: %s", s->log_dir);
	}

	/* Wake up the stormwall daemon if necessary */
	stormwall_msg(s, STORMWALL_ROTATE);

	s->tv.tv_sec = 0;

	/* Open a new one */
	if ( new )
		spool_creat(s);
}

/* rotate all open spools */
/* XXX: Must be called AFTER spool_init_all() */
void spool_rotate_all(void)
{
	struct elog_spool *c;

	for(c = spools.next; c != spools.next; c = c->next)
		spool_rotate(c, 1);
}

/* rotate all spools, only if necessary */
void spool_rotate_check(struct timeval *tv)
{
	struct elog_spool *c;

	for(c = spools.next; c != spools.next; c = c->next) {
		if ( spool_must_rotate(c, tv) )
			spool_rotate(c, 1);
	}
}

/* Rotate a stale logfile. It's racy, but fuck it. */
static int spool_check_old(struct elog_spool *s)
{
	struct {
		struct efile_hdr fh;
		struct elog_pkthdr pkth;
	}buf;
	int fd;

	if ( (fd=open(s->alert_fn, O_RDONLY))<0 )
		return spool_creat(s);

	/* There is a stale alert.elog, so we must now perform
	 * crash recovery. We check it contains valid data and
	 * if so rotate it to a new filename
	 */

	/* Read the date of the first packet and then rotate it */
	if ( read(fd, &buf, sizeof(buf)) < (ssize_t)sizeof(buf) ) {
		goto kill_it;
	}

	/* Check it aint corrupt */
	if ( ntohl(buf.fh.magic) != EF_MAGIC ||
		buf.fh.vers_major != EF_VERS_MAJ ||
		buf.fh.vers_minor != EF_VERS_MIN )
		goto kill_it;

	/* Grab the timestamp of the first entry, we need this
	 * because the filename will be based on this when we
	 * rotate the file. Also if tv_sec is set to zero,
	 * spool_rotate() will think the file is empty and nop */
	s->tv.tv_sec=ntohl(buf.pkth.h.tv_sec);
	s->tv.tv_usec=ntohl(buf.pkth.h.tv_usec);

	/* Rotate it */
	fobuf_newfd(&s->buf, fd);
	spool_rotate(s, 1);
	return 0;

kill_it:
	__fd_close(fd);

	/* If there is no first packet, just unlink it */
	if ( unlink(s->alert_fn) ) {
		mesg(M_WARN, "alert: %s: unlink(): %s",
				s->alert_fn, sys_err());
		return 0;
	}

	mesg(M_WARN, "alert: obliterated %s", s->alert_fn);
	return spool_creat(s);
}

/* finish up with a spool */
void spool_delete(struct elog_spool *s)
{
	if ( !spool_isempty(s) ) {
		/* If we have logged to this file then rotate it */
		spool_rotate(s, 0);
	}else if ( s->buf.fd>=0 ) {
		/* If not, then close and delete it */
		fobuf_newfd(&s->buf, -1);
		if ( unlink(s->alert_fn) )
			mesg(M_CRIT,"alert: unlink(): %s", sys_err());
	}

	/* Close stormwall */
	stormwall_close(s);

	/* Free up all our path information */
	if ( s->alert_fn )
		free(s->alert_fn);

	if ( s->log_dir )
		free(s->log_dir);

	if ( s->fifo_fn )
		free(s->fifo_fn);

	fobuf_detach(&s->buf);

	free(s);
}

static void spool_hup(struct packet *pkt, int l, void *priv)
{
	spool_rotate_all();
}

/* open all alert spools */
void spool_init_all(void)
{
	struct elog_spool *s=spools.next;

	if ( spools.next == (struct elog_spool *)&spools ) {
		mesg(M_WARN, "alert: you have no elog spools specified. "
			"You may lose vital alert information.");
		return;
	}

	while ( s != ((struct elog_spool *)&spools) ) {
		if ( spool_check_old(s)<0 )
			firestorm_exit(EXIT_FAILURE, "alert: error opening alert spool");

		stormwall_open(s);

		s=s->next;
	}

	/* Remove the head from the chain */
	spools.next->prev=spools.prev;
	spools.prev->next=spools.next;

	/* Add post-alert norifier */
	notify(NOTIFY_ALERT, spool_packet, NULL, NULL);
	notify(NOTIFY_HUP, spool_hup, NULL, NULL);
}

/* Send a message to stormwall */
static int stormwall_msg(struct elog_spool *s, int msg)
{
	unsigned char code=msg&0xff;
	int csize=sizeof(code);

	if ( s->stormwall ) {
		if ( write(s->fifo_fd, (char *)&code, csize) < csize ) {
			mesg(M_CRIT, "alert: stormwall: %s", sys_err());
			return -1;
		}
	}

	return 0;
}

/* Open the stormwall communications FIFO */
static void stormwall_open(struct elog_spool *s)
{
	int flags=0;

	if ( s->stormwall==STORMWALL_FAIL ) {
		flags=O_NONBLOCK;
	}else if ( s->stormwall==STORMWALL_WAIT ) {
		mesg(M_DEBUG,"alert: attempting to contact stormwall...");
	}else return;

	if ( (s->fifo_fd=open(s->fifo_fn, O_WRONLY|flags))<0 ) {
		mesg(M_ERR, "%s: open(): %s", s->fifo_fn, sys_err());
		firestorm_exit(EXIT_FAILURE, "alert: unable to contact stormwall");
		return;
	}

	if ( s->stormwall==STORMWALL_WAIT ) {
		mesg(M_DEBUG,"alert: ...SUCCESS!");
	}
}

/* close a stormwall socket */
static void stormwall_close(struct elog_spool *s)
{
	if ( s->fifo_fd<0 )
		return;

	if ( !__fd_close(s->fifo_fd) )
		mesg(M_WARN, "alert: %s: close(): %s",
			s->fifo_fn, sys_err());
}

/* Pad the serialization buffer to 4 bytes and make sure
 * any padding is zeroed.
 */
static inline void pad_buffer(char *buf, size_t *len)
{
	size_t real_len = *len;

	real_len = (*len + 0x3) & ~0x3;

	memset(buf + *len, 0, real_len - *len);

	*len = real_len;
}

/* Convert a alert information in to an elog record */
static int spool_build_packet(struct elog_spool *s, struct alert *a, struct packet *pkt)
{
	static char decode_buf[4096];
	static char sbuf[512];
	struct fvec *iov=s->iov;
	struct elog_dhdr *dh;
	char *ptr=decode_buf;
	char *ptr_end=decode_buf+sizeof(decode_buf);
	off_t tot_len=0;
	int i=0;
	int j;

	/* Construct record header */
	s->ph.h.type=ELOG_ALERT;
	s->ph.h.prio=a->priority;
	s->ph.h.reserved=0;
	s->ph.h.tv_sec=pkt->time.tv_sec;
	s->ph.h.tv_usec=pkt->time.tv_usec;

	/* Construct alert specific bit */
	s->ph.sid=a->sid;
	s->ph.rev=a->rev;
	s->ph.pflags=(pkt->flags & FPMASK_PERSISTANT);
	s->ph.decode_len=0;
	s->ph.gen_len=strlen(a->gen->name)+1;
	s->ph.alert_len=strlen(a->alert)+1;
	s->ph.pkt_len=pkt->len;
	s->ph.pkt_caplen=pkt->caplen;

	/* Record header vector */
	iov[i].ptr=&s->ph;
	iov[i].len=sizeof(s->ph);
	tot_len+=iov[i++].len;

	/* Construct the decode data */
	for(j=0; j<pkt->llen; j++) {
		int slen=0, nlen, real;
		u_int32_t flags;

		if ( pkt->layer[j].proto ) {
			struct layer *l=&pkt->layer[j];

			/* Call plugins to serialise state data */
			if ( l->session && l->proto->serialize ) {
				slen=l->proto->serialize(pkt, j,
					sbuf, sizeof(sbuf));

				if ( slen<0 ) {
					mesg(M_WARN,
						"alert: %s state data too big",
						l->proto->name);
					slen = 0;
				}else{
					pad_buffer(sbuf, &slen);
				}
			}

			real = strlen(l->proto->name)+1;
			nlen = (real + 0x3) & ~0x3;
			flags = htonl(pkt->layer[j].flags);
		}else{
			real=nlen=slen=0;
			flags = 0;
		}

		if ( ptr+nlen+sizeof(*dh)+slen > ptr_end ) {
			mesg(M_CRIT,"elog: decode buffer too small!");
			return -1;
		}

		/* Fill in the header */
		dh=(struct elog_dhdr *)ptr;
		dh->tot_len=(sizeof(*dh)+nlen+slen)>>2;
		dh->name_len=nlen>>2;
		dh->pkt_ofs=htons(pkt->layer[j].h.raw - pkt->base);
		dh->flags=flags;
		ptr+=sizeof(*dh);

		/* Then the protocol name  */
		if ( real ) {
			memcpy(ptr, pkt->layer[j].proto->name, real);
			ptr += real;
			/* make sure padding is zeroed */
			if ( nlen-real ) {
				memset(ptr, 0, nlen-real);
				ptr += nlen-real;
			}
		}

		/* then the session data */
		if ( slen ) {
			memcpy(ptr, sbuf, slen);
			ptr += slen;
		}
	}

	/* decode data vector */
	iov[i].ptr=decode_buf;
	iov[i].len=ptr-decode_buf;

	/* fill in/update record header fields */
	s->ph.decode_len=(iov[i].len>>2);
	tot_len+=iov[i++].len;

	/* raw packet data vector */
	iov[i].ptr=pkt->base;
	iov[i].len=pkt->caplen;
	tot_len+=iov[i++].len;

	/* generator string */
	iov[i].ptr=(char *)a->gen->name;
	iov[i].len=s->ph.gen_len;
	tot_len+=iov[i++].len;

	/* Alert string */
	iov[i].ptr=a->alert;
	iov[i].len=s->ph.alert_len;
	tot_len+=iov[i++].len;

	/* Pad to multiples of 4 bytes (ensures
	 * aligned access to next header if file
	 * is mmap()ed in) */
	if ( tot_len & 3 ) {
		iov[i].ptr=(char *)alert_padding;
		iov[i].len=(4-(tot_len&3));
		tot_len+=iov[i++].len;
	}

	/* Convert to network byte order  */
	s->ph.h.reclen=htonl(tot_len);
	s->ph.h.type=htons(s->ph.h.type);
	s->ph.h.tv_sec=htonl(s->ph.h.tv_sec);
	s->ph.h.tv_usec=htonl(s->ph.h.tv_usec);
	s->ph.sid=htonl(s->ph.sid);
	s->ph.rev=htonl(s->ph.rev);
	s->ph.pflags=htonl(s->ph.pflags);
	s->ph.decode_len=htons(s->ph.decode_len);
	s->ph.pkt_len=htonl(s->ph.pkt_len);
	s->ph.pkt_caplen=htonl(s->ph.pkt_caplen);

	/* phew. this uses 5 or 6 of our I/O vectors, depending on
	 * whether or not we need to pad the end of the record. */
	s->tot_len = tot_len;
	s->io_len = i;
	return 0;
}

/* Select the next spool in the rotation */
static inline struct elog_spool *spool_next(struct timeval *tv)
{
	cur_spool = cur_spool->next;

	/* Rotation filenames contain the timestamp
	 * and serial number of the first packet in
	 * that file */
	if ( !cur_spool->tv.tv_sec ) {
		cur_spool->tv.tv_sec = tv->tv_sec;
		cur_spool->tv.tv_usec = tv->tv_usec;
	}

	return cur_spool;
}

/* Send a packet to the alert spool */
static void spool_packet(struct packet *pkt, int l, void *priv)
{
	struct elog_spool *s = spool_next(&pkt->time);
	int i;

	/* Convert to elog format */
	if ( spool_build_packet(s, pkt->alert, pkt)<0 )
		return;

	/* Check to see if we are about to exceeded max logfile
	 * size or haven't rotated in a while */
	if ( spool_must_rotate(cur_spool, &pkt->time) )
		spool_rotate(cur_spool, 1);

	/* Fast buffered writing */
	for (i=0; i<s->io_len; i++)
		if ( fobuf_write(&s->buf, s->iov[i].ptr, s->iov[i].len) )
			s->size += s->iov[i].len;

	/* We may wish to flush after each packet */
	if ( s->nobuf )
		fobuf_flush(&s->buf);
}
