/*
* This file is part of dotscara
* Copyright (c) 2003 Gianni Tedesco
* Released under the terms of the GNU GPL version 2
*/

#include <stdlib.h>
#include <string.h>
#include <errno.h>
#include <unistd.h>
#include <limits.h>
#include <sys/types.h>
#include <sys/poll.h>

#include <dotscara/dotscara.h>
#include <dotscara/sys.h>
#include <dotscara/fdctl.h>
#include <dotscara/fibuf.h>

static int raw_read(struct fibuf *b, char *buf, size_t len) __nonull(1,2);

static int native_read32(struct fibuf *b, u_int32_t *val)
{
	return raw_read(b, (char *)val, sizeof(*val));
}

static int native_read16(struct fibuf *b, u_int16_t *val)
{
	return raw_read(b, (char *)val, sizeof(*val));
}

static int swap_read32(struct fibuf *b, u_int32_t *val)
{
	if ( !raw_read(b, (char *)val, sizeof(*val)) )
		return 0;
	*val = sys_bswap32(*val);
	return 1;
}

static int swap_read16(struct fibuf *b, u_int16_t *val)
{
	if ( !raw_read(b, (char *)val, sizeof(*val)) )
		return 0;
	*val = sys_bswap16(*val);
	return 1;
}

static const struct fibuf_ops ops_native = {
	.read32 = native_read32,
	.read16 = native_read16,
	.read = raw_read,
};

static const struct fibuf_ops ops_swap = {
	.read32 = swap_read32,
	.read16 = swap_read16,
	.read = raw_read,
};

/**
 * @brief Attach a file descriptor to an input buffer
 * @param b: a fibuf structure to use
 * @param fd: file descriptor to read from
 * @param endian: endianness of the file
 * @param bufsz: size of input buffer to use, 0 is default
 *
 * Attach a file descriptor to an input buffer and ready it for use.
 *
 * @return zero on error, non-zero on success
 */
int __public fibuf_attach(struct fibuf *b, int fd, int endian, size_t bufsz)
{
	const struct fibuf_ops *ops;

	if ( !bufsz )
		bufsz = 4096;

	if ( bufsz > SSIZE_MAX )
		return 0;

	switch(endian) {
	case DSE_BIG_ENDIAN:
#if __BYTE_ORDER == __BIG_ENDIAN
		ops = &ops_native;
#else
		ops = &ops_swap;
#endif
		break;
	case DSE_LITTLE_ENDIAN:
#if __BYTE_ORDER == __LITTLE_ENDIAN
		ops = &ops_native;
#else
		ops = &ops_swap;
#endif
		break;
	case DSE_NATIVE_ENDIAN:
		ops = &ops_native;
		break;
	default:
		return 0;
	}

	b->buf = malloc(bufsz);
	if ( b->buf == NULL )
		return 0;

	b->fd = fd;
	b->read_pos = b->buf;
	b->valid = 0;
	b->buf_sz = bufsz;
	b->eof = 0;
	b->ops = ops;

	return 1;
}

/** fibuf_detach
 * @param b: a fibuf structure to finish up with
 *
 * Free up the buffer and finish with this fibuf.
 */
void __public fibuf_detach(struct fibuf *b)
{
	if ( b->buf )
		free(b->buf);
	b->buf = NULL;
	b->read_pos = NULL;
	b->valid = 0;
	b->buf_sz = 0;
	b->eof = 0;
}

/** fibuf_close
 * @param b: a fibuf strcture to finish up with.
 *
 * Same as fibuf_detach, but we close the file too.
 */
void __public fibuf_close(struct fibuf *b)
{
	if ( b->fd == -1 )
		goto noclose;

	__fd_close(b->fd);

	/* XXX: We don't care about BADF, and EIO is unlikely for a
	 * read-only file descriptor, only problem if the user used
	 * it for writing too, but if they close it via this interface
	 * then they have a bug anyway.
	 */

noclose:
	if ( b->buf )
		free(b->buf);
	b->buf = NULL;
	b->read_pos = NULL;
	b->valid = 0;
	b->buf_sz = 0;
	b->eof = 0;
	b->fd = -1;
}

/** fibuf_newfd
 * @param b: the fibuf structure to modify
 * @param fd: the new file descriptor
 *
 * An optimisation of fbuf_close followed by fbuf_attach,
 */
void __public fibuf_newfd(struct fibuf *b, int fd)
{
	if ( b->fd == -1 )
		goto noclose;

	__fd_close(b->fd);

	/* XXX: We don't care about BADF, and EIO is unlikely for a
	 * read-only file descriptor, only problem if the user used
	 * it for writing too, but if they close it via this interface
	 * then they have a bug anyway.
	 */

noclose:
	b->fd = fd;
	b->valid = 0;
	b->eof = 0;
}

/**
 * @brief slow path for fibuf_read
 * @param b: A fibuf structure to use
 * @param buf: pointer to where you want to read to
 * @param sz: pointer to the length of your buffer
 *
 * Slow path to reading form a buffer, do not call directly, instead use
 * fibuf_read().
 *
 * @return zero on error, non-zero on success. In addition the
 * len parameter is set to the number of bytes actually read.
 */
static int fibuf_read_slow(struct fibuf *b, char *buf, size_t *sz)
{
	size_t len = *sz;
	size_t final;
	size_t siz;
	int ret;

	*sz = 0;

	/* empty the buffer, b->valid may be 0 here */
	memcpy(buf, b->read_pos, b->valid);
	buf += b->valid;
	len -= b->valid;
	*sz += b->valid;
	b->read_pos = b->buf;
	b->valid = 0;

	/* If the remaining data is the same size or bigger than the
	 * buffer, there is no point going through the buffer, lets
	 * save the extra copy and read directly in to the users
	 * supplied buffer.
	 */
	if ( len >= b->buf_sz ) {
		siz = len;
		ret = __fd_read(b->fd, buf, &siz, &b->eof);
		*sz += siz;
		return ret;
	}

	/* fill the buffer up again. */
	siz = b->buf_sz;
	ret = __fd_read(b->fd, b->buf, &siz, &b->eof);
	b->valid = siz;

	/* copy the last bit, left could be 0 again */
	final = ( b->valid < len ) ? b->valid : len;
	memcpy(buf, b->read_pos, final);
	b->read_pos += final;
	b->valid -= final;
	*sz += final;

	return ret;
}

int inline fibuf_read(struct fibuf *b, char *buf, size_t *len)
	__nonull(1,2,3);
int inline fibuf_read(struct fibuf *b, char *buf, size_t *len)
{
	if ( likely(*len <= b->valid) ) {
		memcpy(buf, b->read_pos, *len);
		b->read_pos += *len;
		b->valid -= *len;
		return 1;
	}

	return fibuf_read_slow(b, buf, len);
}

static int raw_read(struct fibuf *b, char *buf, size_t len)
{
	size_t expected = len;

	if ( !fibuf_read(b, buf, &len) )
		return 0;

	if ( len < expected ) {
		if ( unlikely(!fibuf_eof(b)) ) {
			abort();
		}
		return 0;
	}

	return 1;
}
