/*
 * This file is part of Firestorm NIDS
 * Copyright (c) 2004 Gianni Tedesco
 * This program is released under the terms of the GNU GPL version 2
*/

#include <firestorm.h>
#include <dotscara/fobuf.h>
#include <plugin.h>
#include <loader.h>
#include <packet.h>
#include <alert.h>
#include <signature.h>
#include <decode.h>
#include <idx.h>
#include <matcher.h>
#include <i_idx.h>
#include <q_idx.h>
#include <q_filter.h>
#include <field.h>

static struct field *builtins;

struct field *field_builtins(void)
{
	return builtins;
}

static struct detect *val_new(struct matcher *m)
{
	return template_new_node(m->owner->get.val);
}

static struct detect *str_new(struct matcher *m)
{
	return template_new_node_str(m->owner->get.str);
}

static struct detect *str_nocase_new(struct matcher *m)
{
	return template_new_node_str_nocase(m->owner->get.str);
}

static int do_field(struct proto *p, struct field *f)
{
	struct field **list;

	if ( p ) {
		list = &p->fields;
	}else{
		list = &builtins;
	}

	if ( !f->matcher && (!f->get.val || f->type == IDX_INVALID) ) {
		mesg(M_ERR, "field: %s: unset get function / bad type",
				f->name);
		return 0;
	}

	if ( !f->cost || f->matcher )
		goto done;

	f->matcher = calloc(1, sizeof(*f->matcher));
	if ( f->matcher == NULL )
		firestorm_die_error("calloc");

	f->matcher->validate = template_validate;
	f->matcher->insert_default_node = template_insert_default;
	f->matcher->insert_node = template_insert;
	f->matcher->optimize = template_optimize;
	f->matcher->free_priv = free;

	switch( f->type ) {
	case IDX_ORDERED_UINT8:
	case IDX_ORDERED_UINT16:
	case IDX_ORDERED_UINT32:
		f->matcher->new_node = val_new;
		break;
	case IDX_SKUNK:
		f->matcher->new_node = str_new;
		break;
	case IDX_SKUNK_NOCASE:
		f->matcher->new_node = str_nocase_new;
		break;
	default:
		mesg(M_ERR, "field: bad type on field '%s' (%u)",
			f->name, f->type);
		return 0;
	}

done:
	if ( f->matcher ) {
		f->matcher->owner = f;
		matcher_add(f->matcher);
	}

#if 0
	mesg(M_DEBUG, "field: '%s' added to %s with%s matcher with%s field",
			f->name,
			p ? p->name : "builtins",
			f->matcher ? "" : "out",
			(f->get.val && f->type != IDX_INVALID) ? "" : "out");
#endif

	f->next = *list;
	*list = f;

	return 1;
}

void field_add(struct proto *proto, struct field *f)
{
	for(; f->name; f++) {
		if ( !do_field(proto, f) )
			continue;
	}
}

static int alert_get(struct packet *p, int l, struct fvec *vec)
{
	if ( !p->alert )
		return 0;
	vec->ptr = (void *)p->alert->alert;
	vec->len = strlen(p->alert->alert);
	return 1;
}

static int gen_get(struct packet *p, int l, struct fvec *vec)
{
	if ( !p->alert )
		return 0;
	if ( !p->alert->gen )
		return 0;
	vec->ptr = (void *)p->alert->gen->name;
	vec->len = strlen(p->alert->gen->name);
	return 1;
}

static int prio_get(struct packet *p, int l, idx_key_t *val)
{
	if ( !p->alert )
		return 0;
	*val = p->alert->priority;
	return 1;
}

static int sid_get(struct packet *p, int l, idx_key_t *val)
{
	if ( !p->alert )
		return 0;
	*val = p->alert->sid;
	return 1;
}

static int rev_get(struct packet *p, int l, idx_key_t *val)
{
	if ( !p->alert )
		return 0;
	*val = p->alert->rev;
	return 1;
}

static int dsize_get(struct packet *p, int l, idx_key_t *val)
{
	l+=2;

	if ( l >= p->llen ) {
		*val = 0;
		return 1;
	}

	*val = (p->end - p->layer[l].h.raw);

	return 1;
}

static struct field builtin_fields[]={
	{
		.name = "alert",
		.desc = "Alert Description",
		.type = IDX_SKUNK,
		.get = {.str = alert_get},
	},
	{
		.name = "generator",
		.desc = "Generator Name",
		.type = IDX_SKUNK,
		.get = {.str = gen_get},
	},
	{
		.name = "prio",
		.desc = "Alert Priority",
		.type = IDX_ORDERED_UINT8,
		.get = {.val = prio_get},
	},
	{
		.name = "sid",
		.desc = "Signature ID",
		.type = IDX_ORDERED_UINT32,
		.get = {.val = sid_get},
	},
	{
		.name = "rev",
		.desc = "Signature Revision",
		.type = IDX_ORDERED_UINT32,
		.get = {.val = rev_get},
	},
	{
		.name = "dsize",
		.desc = "Payload Size",
		.type = IDX_ORDERED_UINT32,
		.get = {.val = dsize_get},
		.cost = 100,
	},
	{.name = NULL}
};

static void field_register(struct plugin *p, struct field_reg *r)
{
	struct proto *proto;

	for(; r->field; r++) {
		if ( r->proto ) {
			proto = decode_proto(r->proto);
			if ( proto == NULL )
				continue;
		}else{
			proto = NULL;
		}

		field_add(proto, r->field);
	}
}

void field_load(void)
{
	loader_find_syms("firestorm_fields", (loader_cb_t)field_register);
}

void __private field_init(void)
{
	field_add(NULL, builtin_fields);
}
