/*
* This file is part of Firestorm NIDS
* Copyright (c) 2003,2004 Gianni Tedesco
* This program is released under the terms of the GNU GPL version 2
*
* This is the bulk of the GNOME interface code, it's pretty rough
* right now but things should neaten up as we get more of an idea
* how to structure the interface etc. There is noone worth cloning
* after all ;)
*
* TODO:
*  o Properties page for packets
*  o Implement packet context menu
*  o Filter editing dialog
*  o Allow user to customize column layout
*  o Save as / Save Selection as (convert).
*  o Store configuration in GConf
*  o Load elogs in to central DB
*/

#include <firestorm.h>
#include <dotscara/fdctl.h>

#include <unistd.h>
#include <fcntl.h>

#include <gnome.h>
#include <gtkhtml/gtkhtml.h>

#include <alert.h>
#include <target.h>
#include <elog.h>
#include <elog_read.h>
#include <g_pktview.h>
#include <index.h>
#include <idx.h>
#include <q_idx.h>
#include <g_etree_model.h>
#include <g_mesg.h>
#include <g_interface.h>

static EtreeModel *etree;
static GtkWidget *pkt;
static GnomeAppBar *appbar;
GtkWidget *mainwin;

/* Exit callbacks */
static void quit_cb(GtkMenuItem *menuitem, gpointer user_data)
{
	gtk_main_quit();
}

static void delete_cb(void)
{
	gtk_main_quit();
}

static void filter_cb(GtkMenuItem *menuitem, gpointer user_data)
{
	GtkTreeModel *m;
	GtkTreeSelection *s;
	GtkTreeIter i;
	guint sid;

	m = gtk_tree_view_get_model(GTK_TREE_VIEW(user_data));
	s = gtk_tree_view_get_selection(GTK_TREE_VIEW(user_data));
	if ( !gtk_tree_selection_get_selected(s, &m, &i) )
		return;

	gtk_tree_model_get(GTK_TREE_MODEL(etree), &i, ETREE_SID, &sid, -1);
	etree_model_filter_away_sid(etree, sid);
}

static GnomeUIInfo ui_pktmenu[]={
	GNOMEUIINFO_ITEM_STOCK("Filter signature",
				"Filter away alerts caused by this signature",
				filter_cb, GTK_STOCK_REMOVE),
	GNOMEUIINFO_SEPARATOR,
	GNOMEUIINFO_ITEM_STOCK("Export...", "Export packet",
				NULL, GTK_STOCK_CONVERT),
	GNOMEUIINFO_ITEM_STOCK("Delete", "Delete alert",
				NULL, GTK_STOCK_DELETE),
	GNOMEUIINFO_ITEM_STOCK("Properties", "View alert properties",
				NULL, GTK_STOCK_PROPERTIES),
	GNOMEUIINFO_END,
};

static GtkWidget *pkt_menu;

static gboolean tree_clicked(GtkWidget *widget,
				GdkEventButton *event,
				gpointer user_data)
{
	GtkTreeModel *m;
	GtkTreeSelection *s;
	GtkTreeIter i;

	m = gtk_tree_view_get_model(GTK_TREE_VIEW(widget));

	s = gtk_tree_view_get_selection(GTK_TREE_VIEW(widget));
	if ( !gtk_tree_selection_get_selected(s, &m, &i) )
		return FALSE;

	if ( event->button == 3 && event->type == GDK_BUTTON_PRESS ) {
		gnome_popup_menu_do_popup(pkt_menu, NULL, NULL, event,
						widget, widget);
	}

	if ( event->button == 1 && event->type == GDK_2BUTTON_PRESS ) {
		/* Double click */
	}

	return FALSE;
}

/* Selection API */
static int view_fn(struct elog_pkt *p, void *user_data)
{
	return elog_pkt_alert(p, pktview_show, pkt);
}

static void show_pkt(GtkTreeModel *model,
			GtkTreePath *path,
			GtkTreeIter *iter,
			gpointer data)
{
	etree_model_get(ETREE_MODEL(model), iter, view_fn, NULL);
}

static void select_pkt(GtkTreeSelection *sel, gpointer user_data)
{
	gtk_tree_selection_selected_foreach(sel, show_pkt, NULL);
}

/* Open an elog file */
static int display_file(const char *efn)
{
	struct stat est, ist;
	void *efile;
	struct q_db *db;
	int efd, ifd;
	char *ifn;
	int ret = 0;
	int myerror = 1;

	ifn = g_strdup_printf("%s.idx", efn);
	if ( ifn == NULL )
		goto err;

	efd = open(efn, O_RDONLY);
	if ( efd < 0 )
		goto err_free;

try_again:
	ifd = open(ifn, O_RDONLY);
	if ( ifd < 0 ) {
		if ( errno == ENOENT ) {
			mesg(M_INFO, "Generating index '%s'", ifn);
			if ( !index_create(efd, efn) ) {
				myerror = 0;
				goto err_close;
			}
			goto try_again;
		}else{
			goto err_close;
		}
	}

	if ( fstat(efd, &est) )
		goto err_close_idx;
	if ( fstat(ifd, &ist) )
		goto err_close_idx;

	if ( est.st_mtime >= ist.st_mtime ) {
		mesg(M_INFO, "Recreating index '%s' - out of date",
				ifn);
		fd_close(ifd);
		if ( !index_create(efd, efn) ) {
			myerror = 0;
			goto err_close;
		}
		goto try_again;
	}

	db = q_idx_open(ifd);
	if ( db == NULL )
		goto err_close_idx;

	efile = elog_new(efd);
	if ( efile == NULL )
		goto err_close_db;

	if ( etree_model_open(etree, efile, db) ) {
		etree_model_show_all(etree);
		ret = 1;
		goto err_close_idx;
	}

err_close_db:
	q_idx_close(db);
err_close_idx:
	fd_close(ifd);
err_close:
	fd_close(efd);
err_free:
	free(ifn);
err:
	if ( (ret == 0) && myerror ) {
		mesg(M_ERR, "Error opening '%s' - %s", efn, sys_err2(NULL));
	}
	return ret;
}

/* Display the open file selector */
static void open_cb(GtkMenuItem *menuitem, gpointer user_data)
{
	GtkWidget *fc;

	fc = gtk_file_chooser_dialog_new("Select elog file",
					GTK_WINDOW(mainwin),
					GTK_FILE_CHOOSER_ACTION_OPEN,
					GTK_STOCK_CANCEL, GTK_RESPONSE_CANCEL,
					GTK_STOCK_OPEN, GTK_RESPONSE_ACCEPT,
					NULL);

	if ( gtk_dialog_run(GTK_DIALOG(fc)) == GTK_RESPONSE_ACCEPT ) {
		char *fn;
		fn = gtk_file_chooser_get_filename(GTK_FILE_CHOOSER(fc));
		display_file(fn);
		g_free(fn);
	}

	gtk_widget_destroy(fc);
}

/* Unimplemented callbacks */
static void save_cb(GtkMenuItem *menuitem, gpointer user_data){}
static void save_as_cb(GtkMenuItem *menuitem, gpointer user_data){}
static void cut_cb(GtkMenuItem *menuitem, gpointer user_data){}
static void copy_cb(GtkMenuItem *menuitem, gpointer user_data){}
static void paste_cb(GtkMenuItem *menuitem, gpointer user_data){}
static void properties_cb(GtkMenuItem *menuitem, gpointer user_data){}
static void preferences_cb(GtkMenuItem *menuitem, gpointer user_data){}

/* About box */
static void about_cb(GtkMenuItem *menuitem, gpointer user_data)
{
	static GtkWidget *about = NULL;
	gchar *translator_credits = "No translations...";
	gchar *authors[] = {
		"Gianni Tedesco <gianni@scaramanga.co.uk>",
		NULL
	};
	gchar *documenters[] = {
		"Gianni Tedesco <gianni@scaramanga.co.uk>",
		"John Leach <john@johnleach.co.uk>",
		NULL
	};

	if (about != NULL) {
		gtk_window_present(GTK_WINDOW (about));
		return;
	}

	about = gnome_about_new("Firestorm Analyst Console", VERSION,
			"Copyright \xc2\xa9 2002-2004 Gianni Tedesco",
			"http://www.scaramanga.co.uk/firestorm/",
			(const char **)authors, (const char **)documenters,
			strcmp (translator_credits, "translator_credits") != 0 ?
			translator_credits : NULL, NULL);

	g_signal_connect(GTK_OBJECT (about), "destroy",
			G_CALLBACK (gtk_widget_destroyed), &about);

	gtk_widget_show(about);
}

static GnomeUIInfo file1_menu_uiinfo[] = {
	GNOMEUIINFO_MENU_OPEN_ITEM(open_cb, NULL),
	GNOMEUIINFO_MENU_SAVE_ITEM(save_cb, NULL),
	GNOMEUIINFO_MENU_SAVE_AS_ITEM(save_as_cb, NULL),
	GNOMEUIINFO_SEPARATOR,
	GNOMEUIINFO_MENU_EXIT_ITEM(quit_cb, NULL),
	GNOMEUIINFO_END
};

static GnomeUIInfo edit1_menu_uiinfo[] = {
	GNOMEUIINFO_MENU_CUT_ITEM(cut_cb, NULL),
	GNOMEUIINFO_MENU_COPY_ITEM(copy_cb, NULL),
	GNOMEUIINFO_MENU_PASTE_ITEM(paste_cb, NULL),
	GNOMEUIINFO_SEPARATOR,
	GNOMEUIINFO_MENU_PROPERTIES_ITEM(properties_cb, NULL),
	GNOMEUIINFO_SEPARATOR,
	GNOMEUIINFO_MENU_PREFERENCES_ITEM(preferences_cb, NULL),
	GNOMEUIINFO_END
};

static GnomeUIInfo view1_menu_uiinfo[] = {
	GNOMEUIINFO_END
};

static GnomeUIInfo help1_menu_uiinfo[] = {
	GNOMEUIINFO_HELP("gnome-firestorm-console"),
	GNOMEUIINFO_MENU_ABOUT_ITEM(about_cb, NULL),
	GNOMEUIINFO_END
};

static GnomeUIInfo menubar1_uiinfo[] = {
	GNOMEUIINFO_MENU_FILE_TREE(file1_menu_uiinfo),
	GNOMEUIINFO_MENU_EDIT_TREE(edit1_menu_uiinfo),
	GNOMEUIINFO_MENU_VIEW_TREE(view1_menu_uiinfo),
	GNOMEUIINFO_MENU_HELP_TREE(help1_menu_uiinfo),
	GNOMEUIINFO_END
};

static GtkWidget *pkt_tree_init(void)
{
	GtkTreeView *tree;
	GtkTreeViewColumn *col;
	GtkCellRenderer *cell;
	GtkWidget *scroll;

	scroll = gtk_scrolled_window_new(NULL, NULL);
	gtk_scrolled_window_set_policy(GTK_SCROLLED_WINDOW(scroll),
		GTK_POLICY_AUTOMATIC, GTK_POLICY_ALWAYS);

	etree = etree_model_new();

       	tree = GTK_TREE_VIEW(gtk_tree_view_new_with_model(GTK_TREE_MODEL(etree)));
	gtk_tree_view_set_rules_hint(tree, TRUE);
	g_signal_connect(
		G_OBJECT(gtk_tree_view_get_selection(tree)), "changed",
		G_CALLBACK(select_pkt), NULL);
	g_signal_connect(G_OBJECT(tree), "button_press_event",
			G_CALLBACK(tree_clicked), NULL);

	/* Alert column */
	col = gtk_tree_view_column_new();
	cell = gtk_cell_renderer_pixbuf_new();
	gtk_tree_view_column_pack_start(col, cell, FALSE);
	gtk_tree_view_column_set_attributes(col, cell, "pixbuf", ETREE_ICON, NULL);
	cell = gtk_cell_renderer_text_new();
	gtk_cell_renderer_text_set_fixed_height_from_font(
				GTK_CELL_RENDERER_TEXT(cell), 1);
	gtk_tree_view_column_pack_start(col, cell, FALSE);
	gtk_tree_view_column_set_attributes(col, cell, "text", ETREE_PRIO, NULL);
	gtk_tree_view_column_set_title(col, "Priority");
	gtk_tree_view_column_set_resizable(col, TRUE);
	gtk_tree_view_append_column(tree, col);

	col = gtk_tree_view_column_new();
	cell = gtk_cell_renderer_text_new();
	gtk_cell_renderer_text_set_fixed_height_from_font(
				GTK_CELL_RENDERER_TEXT(cell), 1);
	gtk_tree_view_column_pack_start(col, cell, FALSE);
	gtk_tree_view_column_set_attributes(col, cell, "text", ETREE_TIME, NULL);
	gtk_tree_view_column_set_title(col, "Time");
	gtk_tree_view_column_set_resizable(col, TRUE);
	gtk_tree_view_append_column(tree, col);

	col = gtk_tree_view_column_new();
	cell = gtk_cell_renderer_text_new();
	gtk_cell_renderer_text_set_fixed_height_from_font(
				GTK_CELL_RENDERER_TEXT(cell), 1);
	gtk_tree_view_column_pack_start(col, cell, FALSE);
	gtk_tree_view_column_set_attributes(col, cell, "text", ETREE_SRC, NULL);
	gtk_tree_view_column_set_title(col, "Source");
	gtk_tree_view_column_set_resizable(col, TRUE);
	gtk_tree_view_append_column(tree, col);

	col = gtk_tree_view_column_new();
	cell = gtk_cell_renderer_text_new();
	gtk_cell_renderer_text_set_fixed_height_from_font(
				GTK_CELL_RENDERER_TEXT(cell), 1);
	gtk_tree_view_column_pack_start(col, cell, FALSE);
	gtk_tree_view_column_set_attributes(col, cell, "text", ETREE_DST, NULL);
	gtk_tree_view_column_set_title(col, "Destination");
	gtk_tree_view_column_set_resizable(col, TRUE);
	gtk_tree_view_append_column(tree, col);

	col = gtk_tree_view_column_new();
	cell = gtk_cell_renderer_text_new();
	gtk_cell_renderer_text_set_fixed_height_from_font(
				GTK_CELL_RENDERER_TEXT(cell), 1);
	gtk_tree_view_column_pack_start(col, cell, FALSE);
	gtk_tree_view_column_set_attributes(col, cell, "text", ETREE_GEN, NULL);
	gtk_tree_view_column_set_title(col, "Alert");
	gtk_tree_view_column_set_resizable(col, TRUE);
	gtk_tree_view_append_column(tree, col);

	/* Description column */
	col = gtk_tree_view_column_new();
	cell = gtk_cell_renderer_text_new();
	gtk_cell_renderer_text_set_fixed_height_from_font(
				GTK_CELL_RENDERER_TEXT(cell), 1);
	gtk_tree_view_column_pack_start(col, cell, FALSE);
	gtk_tree_view_column_set_attributes(col, cell, "text", ETREE_MSG, NULL);
	gtk_tree_view_column_set_title(col, "Message");
	gtk_tree_view_column_set_resizable(col, TRUE);
	gtk_tree_view_append_column(tree, col);

	gtk_container_add(GTK_CONTAINER(scroll), GTK_WIDGET(tree));

	return scroll;
}

static GtkWidget *pkt_view_new(void)
{
	GtkWidget *scroll;

	scroll = gtk_scrolled_window_new(NULL, NULL);
	gtk_scrolled_window_set_policy(GTK_SCROLLED_WINDOW(scroll),
		GTK_POLICY_AUTOMATIC, GTK_POLICY_ALWAYS);

	pkt = pktview_new();

	gtk_container_add(GTK_CONTAINER(scroll), pkt);

	return scroll;
}

void g_interface_mainwin(void)
{
	GtkWidget *toolbar;
	GtkWidget *vpane;
	GtkWidget *tabs;
	GtkWidget *main_box;

	g_mesg_init();

	/* Create main window */
	mainwin = gnome_app_new("gnome-firestorm-console",
		"Firestorm Analyst Console");
	gtk_signal_connect(GTK_OBJECT(mainwin),
		"delete-event", delete_cb, NULL);
	gtk_window_set_default_size(GTK_WINDOW(mainwin), 800, 600);

	/* Status bar */
	appbar = (GnomeAppBar *)gnome_appbar_new(TRUE, TRUE,
			GNOME_PREFERENCES_USER);
	gnome_app_set_statusbar(GNOME_APP(mainwin), GTK_WIDGET(appbar));

	/* Stick the menus in */
	gnome_app_create_menus(GNOME_APP(mainwin), menubar1_uiinfo);
	gnome_app_install_menu_hints(GNOME_APP(mainwin), menubar1_uiinfo);

	pkt_menu = gnome_popup_menu_new(ui_pktmenu);
	gnome_app_install_menu_hints(GNOME_APP(mainwin), ui_pktmenu);

	/* Add the toolbar  */
	toolbar = gtk_toolbar_new();
	gnome_app_add_toolbar(GNOME_APP(mainwin),
			GTK_TOOLBAR(toolbar),
			"toolbar",
			BONOBO_DOCK_ITEM_BEH_EXCLUSIVE,
			BONOBO_DOCK_TOP,
			1, 0, 0);

	/* The three toolbar buttons */
	gtk_toolbar_insert_stock(GTK_TOOLBAR(toolbar),
		"gtk-open", NULL, NULL,
		GTK_SIGNAL_FUNC(open_cb), NULL, -1);
	gtk_toolbar_insert_stock(GTK_TOOLBAR(toolbar),
		"gtk-save", NULL, NULL,
		GTK_SIGNAL_FUNC(NULL), NULL, -1);
	gtk_toolbar_insert_stock(GTK_TOOLBAR(toolbar),
		"gtk-find", NULL, NULL,
		GTK_SIGNAL_FUNC(NULL), NULL, -1);

	vpane = gtk_vpaned_new();

	/* Packet view */
	main_box = pkt_tree_init();

	/* Tab sheet */
	tabs = gtk_notebook_new();
	gtk_notebook_append_page(GTK_NOTEBOOK(tabs),
		pkt_view_new(),
		gtk_label_new("Details"));
	gtk_notebook_append_page(GTK_NOTEBOOK(tabs),
		g_mesg_view_new(),
		gtk_label_new("Messages"));

	gtk_paned_add1(GTK_PANED(vpane), main_box);
	GTK_PANED(vpane)->child1_resize = TRUE;
	GTK_PANED(vpane)->child1_shrink = TRUE;

	gtk_paned_add2(GTK_PANED(vpane), tabs);
	GTK_PANED(vpane)->child2_resize = FALSE;
	GTK_PANED(vpane)->child2_shrink = FALSE;

	gnome_app_set_contents(GNOME_APP(mainwin), vpane);

	gtk_widget_show_all(mainwin);
}
