/*
 * This file is part of Firestorm NIDS
 * Copyright (c) 2003,2004 Gianni Tedesco
 * Released under the terms of the GNU GPL version 2
 *
 * This file represents the index creation API.
*/

#include <firestorm.h>
#include <errno.h>
#include <unistd.h>
#include <fcntl.h>

#include <dotscara/fobuf.h>
#include <index.h>
#include <elog.h>
#include <alert.h>
#include <signature.h>
#include <decode.h>
#include <target.h>
#include <elog_read.h>
#include <idx.h>
#include <i_idx.h>
#include <q_idx.h>
#include <q_filter.h>
#include <field.h>

static int index_insert(struct field *f, struct packet *p,
			int l, idx_ofs_t ofs)
{
	struct i_fieldref *fr = f->ornament;
	struct fvec vec;
	idx_key_t key;

	if ( l != -1 )
		l -= f->ofs;

	if ( l < -1 || l > p->llen )
		return 1;

	switch(f->type) {
	case IDX_INVALID:
		return 1;
	case IDX_ORDERED_UINT8:
	case IDX_ORDERED_INT8:
	case IDX_ORDERED_UINT16:
	case IDX_ORDERED_INT16:
	case IDX_ORDERED_UINT32:
	case IDX_ORDERED_INT32:
		if ( !f->get.val(p, l - f->ofs, &key) )
			return 1;
		return fr->type->insert(fr, key, ofs);
	case IDX_SKUNK:
	case IDX_SKUNK_NOCASE:
		if ( !f->get.str(p, l - f->ofs, &vec) )
			return 1;
		return fr->type->insert_str(fr, vec.ptr, vec.len, ofs);
	default:
		return 0;
	}
}

static int idx_target_fn(struct packet *p, void *priv)
{
	struct field *f;
	size_t cur_ofs = p->u.elog->ofs;
	int i;

	/* Do non-layer stuff */
	for(f=field_builtins(); f; f=f->next) {
		if ( !index_insert(f, p, -1, cur_ofs) )
			goto err;
	}

	/* Do the per-layer stuff */
	for(i=0; i < p->llen; i++) {
		if ( p->layer[i].proto == NULL )
			continue;

		for(f=p->layer[i].proto->fields; f; f=f->next) {
			if ( !index_insert(f, p, i, cur_ofs) )
				goto err;
		}
	}

	return 1;

err:
	mesg(M_ERR, "Error adding '%s' field to index: %s",
			f->name, sys_err2(NULL));
	return 0;
}

static int cbfn_idx(struct elog_pkt *p, void *data)
{
	return elog_pkt_alert(p, idx_target_fn, data);
}

static int cbfn_field(struct field *f, void *idx)
{
	if ( f->type == IDX_INVALID )
		return 1;

	f->ornament = i_idx_add_field(idx, f->type, f->name);
	if ( f->ornament == NULL )
		return 0;

	return 1;
}

int index_create(int fd, const char *fn)
{
	struct i_idx *idx;
	void *ep;
	char nbuf[strlen(fn) + 9];
	char fbuf[strlen(fn) + 5];
	int ofd;
	const char *err_fn = fn;
	const char *err_func = NULL;

	/* build filenames */
	snprintf(nbuf, sizeof(nbuf), "%s.idx.tmp", fn);
	snprintf(fbuf, sizeof(fbuf), "%s.idx", fn);

	/* Open the input file */
	ep = elog_new(fd);
	if ( ep == NULL ) {
		err_func = "elog_new";
		goto err;
	}

	/* Open the output file */
	ofd = open(nbuf, O_WRONLY|O_TRUNC|O_CREAT, 0640);
	if ( ofd < 0 ) {
		err_func = "open";
		err_fn = nbuf;
		goto err_free;
	}

	idx = i_idx_new();
	if ( idx == NULL ) {
		err_func = "i_idx_new";
		err_fn = nbuf;
		goto err_free;
	}

	/* Add each field to the DB */
	if ( field_foreach(cbfn_field, idx) != NULL ) {
		err_func = "i_idx_add_field";
		err_fn = nbuf;
		goto err_idx;
	}

	/* Run through the file */
	elog_set_alert_fn(ep, cbfn_idx);
	if ( elog_run(ep, NULL) <= 0 ) {
		err_func = "elog_run";
		goto err_idx;
	}

	/* Write the index file out, this syncs and closes the fd */
	if ( !i_idx_write_fd(idx, ofd) ) {
		err_func = "i_idx_write_fd";
		err_fn = nbuf;
		goto err_idx;
	}

	/* Free everything up */
	i_idx_free(idx);

	/* Rename temporary file over real one */
	if ( rename(nbuf, fbuf) ) {
		err_func = "rename";
		err_fn = nbuf;
		goto err_unlink;
	}

	elog_free(ep);

	return 1;

err_idx:
	i_idx_free(idx);
err_unlink:
	if ( unlink(nbuf) ) {
		if ( errno == EINTR )
			goto err_unlink;
		mesg(M_ERR, "%s: unlink(): %s", nbuf, sys_err());
	}
err_free:
	elog_free(ep);
err:
	mesg(M_ERR, "%s: %s: %s", err_fn, err_func, sys_err2(NULL));
	return 0;
}
