/*
* This file is part of dotscara
* Copyright (c) 2003 Gianni Tedesco
* Released under the terms of the GNU GPL version 2
*/

#include <stdlib.h>
#include <string.h>

#include <dotscara/dotscara.h>
#include <dotscara/mpool.h>

/** mpool_new
 * @param m: an mpool structure to use
 * @param obj_size: size of objects to allocate
 * @param slab_size: size of slabs (set to zero for auto)
 *
 * Creates a new empty memory pool descriptor with the passed values
 * set. The resultant mpool has no alignment requirement set.
 *
 * Return value: zero on error, non-zero for success
 */
int __public mpool_new(struct mpool *m, size_t obj_size, size_t slab_size)
{
	/* quick sanity checks */
	if ( obj_size == 0 )
		return 0;

	if ( obj_size < sizeof(void *) )
		obj_size=sizeof(void *);

	if ( slab_size && slab_size < obj_size )
		return 0;

	m->obj_size = obj_size;

	if ( slab_size ) {
		m->slab_size = slab_size;
	}else{
		/* XXX: totally arbitrary... */
		m->slab_size = 8192;
	}

	m->slabs=NULL;

	return 1;
}

/** __mpool_alloc_slow
 * @param m: a valid mpool structure returned from mpool_new()
 *
 * Allocate a new object, returns NULL if out of memory. Note that this
 * is the slow path, the fast path for common case (no new allocation
 * needed) is handled in the inline fucntion mpool_alloc() in mpool.h.
 *
 * Return value: a new object
 */
void __public *__mpool_alloc_slow(struct mpool *m)
{
	size_t num;
	struct mpool_hdr *h;
	void *ptr, *ret;
	
	num = (m->slab_size - sizeof(*h)) / m->obj_size;
	h = ptr = malloc(m->slab_size);
	if ( h == NULL )
		return NULL;

	/* Set first object */
	h->next_obj = ptr + sizeof(*h) + m->obj_size;
	ret = ptr + sizeof(*h);

	/* prepend to slab list */
	h->next = m->slabs;
	m->slabs = h;

	return ret;
}

/** mpool_free
 * @param m: a valid mpool structure returned from mpool_new()
 *
 * Frees up all allocated memory from the #mpool and resets all members
 * to invalid values.
 */
void __public mpool_free(struct mpool *m)
{
	struct mpool_hdr *h, *f;

	for(h=m->slabs; (f=h) ; free(f))
		h=h->next;

	memset(m, 0, sizeof(*m));
}

/** mpool_destroy
 * @param m: a valid mpool structure returned from mpool_new()
 * @param dtor: A destructor function to call
 *
 * Same as mpool_free() but calls a destructor on every
 * allocated object.
 */
void __public mpool_destroy(struct mpool *m, void(*dtor)(void *))
{
	struct mpool_hdr *h, *f;
	void *p;

	for(h=m->slabs; (f=h); h=h->next ) {
		for(p=f->data; p < f->next_obj; p += m->obj_size) {
			(*dtor)(p);
		}
	}

	memset(m, 0, sizeof(*m));
}
