/*
* This file is part of Firestorm NIDS
* Copyright (c) 2003,2004 Gianni Tedesco
* Released under the terms of the GNU GPL version 2
*
* This code provides an interface for setting up stored queries, which
* you can then simply run against an index file. It can either take
* textual queries or be programmed. Queries can only be stored in
* memory currently.
*
* TODO:
*  o Fix the query scheduler
*  o Support adding constraints to queries at any time
*  o Implement merge_union() for OR queries
*  o Implement merge_difference() for NOT queries
*  o SMP support for query scheduler
*/

#include <firestorm.h>

#include <sys/stat.h>
#include <sys/mman.h>
#include <unistd.h>

#include <dotscara/fobuf.h>
#include <strtouint.h>
#include <alert.h>
#include <idx.h>
#include <i_idx.h>
#include <q_idx.h>
#include <q_ordered.h>
#include <q_skunk.h>
#include <q_filter.h>
#include <q_type.h>

/* Must be a power of two */
#define PREALLOC_CHUNK 8
#define DEFAULT_PREALLOC 32

/* merge_intersect:
 * @o: A results array
 * @alen: number of elements in results array
 * @r: A database result
 *
 * Merge sorts a database lookup result in to the supplied array.
 *
 * Return value: number of results
 */
static int merge_intersect(idx_val_t *o, unsigned int alen, struct q_res *r)
{
	idx_ofs_t ia, io;
	idx_val_t *a=o;
	idx_val_t *b[r->num_keys];
	idx_val_t *eb[r->num_keys];
	idx_ofs_t i;

	for(ia=i=0; i<r->num_keys; i++) {
		b[i] = &r->val[ia];
		ia += r->node[i].num;
		eb[i] = &r->val[ia];
	}

	for(io=ia=0; ia<alen; ia++) {
		for(i=0; i<r->num_keys; i++) {
			while ( b[i] < eb[i] &&
				a[ia] > *b[i] ) {
				b[i]++;
			}

			if ( b[i] >= eb[i] )
				continue;

			while ( a[ia] == *b[i] ) {
				if ( !(io && o[io-1] == a[ia]) ) {
					o[io] = a[ia];
					io++;
				}
				b[i]++;
				if ( b[i] >= eb[i] )
					break;
			}
		}
	}

	return io;
}

/* merge_sort:
 * @o: Array to sort in to
 * @r: A database result
 *
 * Merges a single database query result in to an empty array.
 *
 * Return value: number of results
 */
static int merge_sort(idx_val_t *o, struct q_res *r)
{
	idx_val_t *b[r->num_keys];
	idx_val_t *eb[r->num_keys];
	idx_ofs_t io, i;

	/* Calculate start and end pointers for eaach stream to merge */
	for(io=i=0; i < r->num_keys; i++) {
		b[i] = &r->val[io];
		eb[i] = &r->val[io + r->node[i].num];
		io += r->node[i].num;
	}

	/* Do the n-way intersecting merge against o*/
	for(io=0; io < r->num_vals; ) {
		idx_val_t lval = *b[0];
		int lowest = 0;

		for(i=0; i < r->num_keys; i++) {
			if ( b[i] >= eb[i] )
				continue;

			if ( *b[i] <= lval ) {
				lval = *b[i];
				lowest = i;
			}
		}

		b[lowest]++;

		if ( io && o[io-1] == lval )
			continue;
		o[io++] = lval;
	}

	return io;
}

/* q_filter_assure:
 * @qf: A valid q_filter structure
 * @num: Number of constraints to assure
 *
 * Make sure that the constraint array is big enough for num
 * elements, allocating more space if necessary.
 *
 * Return value: 1 on success, 0 on failure
 */
static int q_filter_assure(struct q_filter *qf, size_t num)
{
	size_t sz;
	void *new;

	num = (num + (PREALLOC_CHUNK-1)) & ~(PREALLOC_CHUNK-1);
	sz = num * sizeof(*qf->constraint);

	if ( qf->num_alloc >= num )
		return 1;

	new = realloc(qf->constraint, sz);
	if ( new ) {
		qf->num_alloc = sz / sizeof(*qf->constraint);
		qf->constraint = new;
	}

	return (num >= qf->num_alloc);
}

/* q_filter_new:
 * @prealloc: Number of constraints to preallocate
 *
 * Allocate a new q_filter structure.
 *
 * Return value: NULL on error, otherwise a pointer to the new structure.
 */
struct q_filter *q_filter_new(size_t prealloc)
{
	struct q_filter *ret;

	ret = calloc(1, sizeof(*ret));
	if ( ret == NULL )
		return 0;

	/* Allocate the first set of constraints, if prealloc is set */
	if ( prealloc )
		q_filter_assure(ret, prealloc);

	/* If it isn't set or fails, go back to the default */
	if ( prealloc == 0 || prealloc > DEFAULT_PREALLOC )
		q_filter_assure(ret, DEFAULT_PREALLOC);

	/* Never less than the minimum possible */
#if PREALLOC_CHUNK < DEFAULT_PREALLOC
	q_filter_assure(ret, PREALLOC_CHUNK);
#endif

	if ( ret->num_alloc == 0 ) {
		free(ret);
		return NULL;
	}

	return ret;
}

/* q_filter_free:
 * @qf: A valid q_filter structure
 *
 * Frees up all allocated memory to do with the q_filter.
 */
void q_filter_free(struct q_filter *qf)
{
	if ( qf->constraint )
		free(qf->constraint);
	if ( qf->result )
		free(qf->result);
	if ( qf->filter )
		free(qf->filter);
	if ( qf->query )
		free(qf->query);
	free(qf);
}

/* q_filter_free_result:
 * @qf: A valid q_filter structure
 *
 * free up the results after calling running a query with
 * q_filter_run().
 */
void q_filter_free_result(struct q_filter *qf)
{
	if ( qf->result ) {
		free(qf->result);
		qf->result = NULL;
		qf->num_recs = 0;
	}
}

/* q_filter_constraint:
 * @qf: A valid q_filter structure
 * @field: Name of the field
 * @op: operator
 * @val: value
 *
 * Add a constraint to a q_filter structure. This is the
 * programmable query interface.
 *
 * Return value: Zero on error, non-zero for success.
 */
int q_filter_constraint(struct q_filter *qf, const char *field,
			int op, const char *val)
{
	int i = qf->num_constraints;
	struct q_constraint *c;
	struct q_type *type = NULL;

	/* Validate input */
	if ( field == NULL || op >= QF_OP_MAX )
		return 0;

	type = q_type_get(field, val);
	if ( type == NULL )
		return 0;

	if ( !q_filter_assure(qf, i + 1) )
		return 0;

	c = &qf->constraint[i];
	c->field = field;
	c->operator = op;
	c->type = type;
	c->type->validate(val, c);

	qf->num_constraints++;
	return 1;
}

/* qsort constraints by name */
static int cmp_name(const void *a, const void *b)
{
	const struct q_constraint *A=a, *B=b;
	return strcmp(A->field, B->field);
}

/* field_intersect_int:
 * @f: a valid db_filter structure
 * @op: Operator
 * @key: min part of operand range
 * @max: max part of operand range
 *
 * Add queries together to form an intersection for that one field.
 *
 * Return value: Zero on error, otherwise non-zero
 */
static int field_intersect_int(struct q_db_filter *f, int op,
				idx_key_t key, idx_key_t max)
{
	if ( key > max )
		return 0;
	if ( (key != max) && (op != QF_OP_EQ) )
		return 0;

	switch(op) {
	case QF_OP_EQ:
		if ( key < f->u.r.min || max > f->u.r.max )
			return 0;
		f->u.r.min = key;
		f->u.r.max = max;
		break;
	case QF_OP_LT:
		key--;
		if ( f->u.r.min > key )
			return 0;
		f->u.r.max = key;
		break;
	case QF_OP_GT:
		key++;
		if ( f->u.r.max < key )
			return 0;
		f->u.r.min = key;
		break;
	case QF_OP_LE:
		if ( f->u.r.min > key )
			return 0;
		f->u.r.max = key;
		break;
	case QF_OP_GE:
		if ( f->u.r.max < key )
			return 0;
		f->u.r.min = key;
		break;
	case QF_OP_NE:
	default:
		mesg(M_ERR, "%s: int: %i operator not supported\n",
				f->field, op);
		return 0;
	}

	return 1;
}

static int field_intersect_str(struct q_db_filter *f, int op, struct fvec *key)
{
	if ( op != QF_OP_EQ ) {
		mesg(M_ERR, "%s: str: %i operator not supported\n",
				f->field, op);
		return 0;
	}

	if ( !f->u.vec.ptr ) {
		f->u.vec.ptr = key->ptr;
		f->u.vec.len = key->len;
		return 1;
	}

	if ( f->u.vec.len == key->len &&
			!memcmp(f->u.vec.ptr, key->ptr, key->len) )
			return 1;

	return 0;
}

/* q_filter_prepare:
 * @qf: A valid q_filter structure
 *
 * Prepare a filter for running queries. Must be called before
 * q_filter_run() if you are using the programmable interface.
 *
 * Return value: Zero on error, otherwise non-zero.
 */
int q_filter_prepare(struct q_filter *qf)
{
	const char *cur = NULL;
	struct q_db_filter *f;
	size_t num;
	unsigned int i, x;
	int ret = 0;

	if ( qf->num_constraints == 0 )
		return 0;

	qsort(qf->constraint, qf->num_constraints,
		sizeof(*qf->constraint), cmp_name);

	/* Count the number of different fields */
	for(i=num=0; i < qf->num_constraints; i++) {
		const char *this = qf->constraint[i].field;

		if ( cur == NULL || strcmp(cur, this) ) {
			cur = this;
			num++;
		}
	}

	/* Allocate the new filter structure */
	f = calloc(num, sizeof(*f));
	if ( f == NULL )
		return 0;

	/* Pre-process the information */
	for(x=i=0; i < num; i++) {
		f[i].field = qf->constraint[x].field;
		f[i].type = qf->constraint[x].type;
		switch( f[i].type->type ) {
		case QFT_INT:
			f[i].u.r.min = IDX_KEY_MIN;
			f[i].u.r.max = IDX_KEY_MAX;
			break;
		case QFT_STRING:
			f[i].u.vec.ptr = NULL;
			f[i].u.vec.len = 0;
			break;
		}

		while( x < qf->num_constraints &&
			!strcmp(f[i].field, qf->constraint[x].field) ) {
			int r;

			if ( f[i].type != qf->constraint[x].type )
				goto out;

			switch ( f[i].type->type ) {
			case QFT_INT:
				r = field_intersect_int(&f[i],
						qf->constraint[x].operator,
						qf->constraint[x].u.op_int.min,
						qf->constraint[x].u.op_int.max);
				break;
			case QFT_STRING:
				r = field_intersect_str(&f[i],
						qf->constraint[x].operator,
						&qf->constraint[x].u.op_vec);
				break;
			default:
				goto out;
			}

			if ( !r ) {
				qf->noresult = 1;
				ret = 1;
				goto out;
			}
			x++;
		}

	}

	qf->filter = f;
	qf->num_filter = num;
	ret = 1;

	/* Get rid of the primary data structures */
out:
	free(qf->constraint);
	qf->constraint = NULL;
	qf->num_constraints = 0;
	qf->num_alloc = 0;

	return ret;
}

/* qsort results by number of values */
static int cmp_res(const void *aa, const void *bb)
{
	const struct q_res *a=aa;
	const struct q_res *b=bb;

	if ( a->num_vals < b->num_vals ) return -1;
	if ( a->num_vals > b->num_vals ) return 1;
	return 0;
}

/* q_filter_run:
 * @qf: A valid, prepared q_filter structure
 * @fd: A readable file descriptor for a database
 *
 * Runs the given query against the given index file. Returns
 * results in to the @qf structure.
 *
 * Return value: Zero on error, non-zero otherwise
 */
int q_filter_run(struct q_filter *qf, struct q_db *db)
{
	size_t num = qf->num_filter;
	struct q_res res[num];
	idx_val_t *ofs;
	size_t olen=0;
	unsigned int i;
	int ret=0;

	if ( num == 0 )
		return 0;

	/* This can happen if we figured out during preparation
	 * that the constraints conflict meaning there are zero
	 * results (ie:  x > 10 && x < 10)
	 */
	if ( qf->noresult )
		return 1;

	/* Perform all the queries */
	for(i=0; i < num; i++) {
		struct q_idx *q;

		q = q_idx_get(db, qf->filter[i].field);
		if ( q == NULL )
			goto err;

		if ( !q_type_compatible(qf->filter[i].type, q->type) )
			goto err;

		switch(q->type) {
		case IDX_ORDERED_UINT32:
			if ( !q_ordered_range(q, &res[i],
				qf->filter[i].u.r.min,
				qf->filter[i].u.r.max) )
				goto err;
			break;
		case IDX_SKUNK:
			if ( !q_skunk_query(q, &res[i], &qf->filter[i].u.vec) )
				goto err;
			break;
		case IDX_SKUNK_NOCASE:
			if ( !q_skunk_nocase(q, &res[i], &qf->filter[i].u.vec) )
				goto err;
			break;
		default:
			mesg(M_ERR, "%s: type %i not supported",
				q->name, q->type);
		}

		if ( res[i].num_keys == 0 ) {
			ret = 1;
			goto err;
		}
	}

	/* Sort results by number of keys */
	qsort(res, num, sizeof(*res), cmp_res);

	/* Seeing as this is an intersection operation, the maximum
	 * size of the resulting set is the size of the smallest set.
	 */
	ofs = malloc(sizeof(*ofs) * res[0].num_vals);
	if ( ofs == NULL )
		goto err;

	/* Merge result 0 with itself */
	olen = merge_sort(ofs, res);

	/* Merge all other results */
	for(i=1; i < num; i++) {
		olen = merge_intersect(ofs, olen, &res[i]);
		if ( olen == 0 )
			break;
	}

	/* Shrink set to real number of results */
	realloc(ofs, olen * sizeof(*ofs));
	if ( !olen )
		ofs = NULL;

	ret = 1;
	qf->result = ofs;
	qf->num_recs = olen;
err:
	q_idx_close(db);
	return ret;
}

/* q_filter_operator:
 * @op: A string
 *
 * Returns an operator code given a string represenetation
 *
 * Return value: -1 on error, a valid operator code otherwise.
 */
static int q_filter_operator(const char *op)
{
	static struct {
		const char *str;
		int val;
	}ops[]={
		{"==", QF_OP_EQ},
		{"=",  QF_OP_EQ},
		{">",  QF_OP_GT},
		{"<",  QF_OP_LT},
		{">=", QF_OP_GE},
		{"<=", QF_OP_LE},
		{"!=", QF_OP_NE},
	};

	int num = sizeof(ops) / sizeof(*ops);
	int i;

	for(i=0; i < num; i++) {
		if ( !strcmp(ops[i].str, op) )
			return ops[i].val;
	}

	return -1;
}

/* q_filter_query:
 * @qf: A valid q_filter structure
 * @q: A query string
 *
 * The non-programmable interface, this takes the textual query and
 * sets up the q_filter appropriately. It also calls q_filter_prepare().
 *
 * Return value: Zero on error, non-zero otherwise.
 */
int q_filter_query(struct q_filter *qf, const char *q)
{
	int state = 0;
	char opbuf[3];
	char *fn=NULL, *val=NULL;
	int seenwhite=0;
	char *cur;
	int oi=0;

	qf->query = strdup(q);
	if ( qf->query == NULL )
		return 0;

	for(cur=qf->query; ; cur++) {
		if ( *cur == '\0' && state != 4 )
			return 0;

		switch(state) {
		case 0:
			if ( !isspace(*cur) ) {
				fn = cur;
				seenwhite = 0;
				state = 1;
			}
			break;
		case 1:
			if ( *cur == '=' || *cur == '>' ||
				*cur == '<' || *cur == '!' ) {
				oi=0;
				opbuf[oi++] = *cur;
				*cur = '\0';
				state = 2;
			}else if ( isspace(*cur) ) {
				*cur = '\0';
				seenwhite = 1;
			}else if ( seenwhite ){
				return 0;
			}
			break;
		case 2:
			if ( *cur != '=' && *cur != '>' &&
				*cur != '<' && *cur != '!' ) {
				if ( oi == 0 )
					return 0;
				opbuf[oi] = '\0';
				state = 3;
				cur--;
			}else{
				opbuf[oi++] = *cur;
				if ( oi > 2 )
					return 0;
			}
			break;
		case 3:
			if ( !isspace(*cur) ) {
				val = cur;
				state = 4;
			}
			break;
		case 4:
			if ( isspace(*cur) || *cur == '\0' ) {
				int op;
				int end;

				if ( *cur == '\0' ) {
					end = 1;
				}else{
					end = 0;
					*cur = '\0';
				}

				op = q_filter_operator(opbuf);
				if ( op < 0 )
					return 0;

				if ( !q_filter_constraint(qf, fn, op, val) )
					return 0;

				state = 0;
				if ( end )
					goto done;
			}
			break;
		}
	}

done:
	return q_filter_prepare(qf);
}
