/*
* This file is part of Firestorm NIDS
* Copyright (c) 2002,2003,2004 Gianni Tedesco
* This program is released under the terms of the GNU GPL version 2
*
* This code handles alerts, it mainly just calls functions in
* elog_write.c.
*
* Functions:
*  o alert_hup() - rotate log on demand
*  o cb_dir() - set log spool directory path
*  o cb_stormwall() - set stormwall mode
*  o alert_conf_hook() - called for each output directive
*  o alert_conf_go() - open all log spools
*  o alert_hook() - the hook called for alerts
*/
#include <firestorm.h>
#include <signal.h>
#include <sys/stat.h>
#include <sys/socket.h>
#include <fcntl.h>
#include <netinet/in.h>

#include <sensor.h>
#include <alert.h>
#include <elog.h>
#include <elog_write.h>
#include <signature.h>
#include <decode.h>
#include <args.h>
#include <capture.h>
#include <target.h>

static unsigned int timeout = 0;

static int cb_dir(struct arg *a, void *priv)
{
	struct elog_spool *s=priv;
	static char sw[]=".stormwall";
	int l;

	if ( !s )
		return 0;

	if ( !a->val.v_str )
		return 0;

	if ( *a->val.v_str==0 )
		return 0;

	if ( s->rotate_fn )
		return 0;

	l=strlen(a->val.v_str);

	/* We add a slash unconditionally */
	if ( a->val.v_str[l-1]=='/' ) {
		a->val.v_str[l-1]=0;
		l-=1;
	}

	/* Allocate for the log dir */
	if ( !(s->log_dir=malloc(l + ALERT_FNLEN + 2)) ) {
		return 0;
	}

	s->rotate_fn=s->log_dir + l + 1;
	sprintf(s->log_dir, "%s/", a->val.v_str);

	/* Allocate enough space for the whole filename */
	if ( !(s->alert_fn=malloc(l + strlen(ALERT_FN) + 2)) ) {
		free(s->log_dir);
		return 0;
	}

	/* Figure out stormwall fifo filename */
	if ( !(s->fifo_fn=malloc(l+sizeof(sw)+2)) ) {
		free(s->alert_fn);
		free(s->log_dir);
		return 0;
	}

	/* Generate the filenames */
	sprintf(s->alert_fn, "%s/%s", a->val.v_str, ALERT_FN);
	sprintf(s->fifo_fn, "%s/%s", a->val.v_str, sw);

	return 1;
}

static int cb_stormwall(struct arg *a, void *priv)
{
	struct elog_spool *s=priv;

	if ( !s )
		return 0;

	if ( !strcmp("none", a->val.v_str) ) {
		s->stormwall=STORMWALL_NONE;
	}else if ( !strcmp("wait", a->val.v_str) ) {
		s->stormwall=STORMWALL_WAIT;
	}else if ( !strcmp("fail", a->val.v_str) ) {
		s->stormwall=STORMWALL_FAIL;
	}else{
		mesg(M_ERR, "output: stormwall can be one of 'none',"
			"'wait' or 'fail'");
		return 0;
	}

	return 1;
}

/* euclids greatest common divisor */
static inline unsigned int gcd(unsigned int n, unsigned int d)
{
	unsigned int r;

	if ( n == 0 )
		return d;

	if ( n < d ) {
		unsigned int tmp;
		tmp = d;
		d = n;
		n = tmp;
	}
	
	while ( d > 0 ) {
		r = n % d;
		n = d;
		d = r;
	}

	return n;
}

/* called once for each 'output' directive on the config file */
void alert_conf_hook(const char *args)
{
	struct elog_spool *s;
	size_t buf_sz=SPOOL_DEFAULT_BUFFER;
	static struct arg a[]={
		{"dir", ARGTYPE_STRING, cb_dir},
		{"stormwall", ARGTYPE_STRING, cb_stormwall},
		{"size", ARGTYPE_POFF, NULL},
		{"minutes", ARGTYPE_PUINT, NULL},
		{"buf", ARGTYPE_PBYTES, NULL},
		{NULL, ARGTYPE_NOP, NULL}
	};

	/* Allocate the spool */
	if ( !(s=spool_new()) )
		firestorm_exit(EXIT_FAILURE, "alert: spool_new(): %s", sys_err());

	/* setup the arg pointers */
	a[2].val.vp_off = &s->max_bytes;
	a[3].val.vp_uint = &s->max_time;
	a[4].val.vp_bytes = &buf_sz;

	/* parse the damn things */
	if ( args ) {
		switch ( args_parse(a, args, s) ) {
		case -1:
			mesg(M_ERR, "output: parse error: %s", args);
		case 0:
			firestorm_exit(EXIT_FAILURE,"output: failed");
		default:
			break;
		}
	}

	/* Must specify a log dir */
	if ( !s->alert_fn )
		firestorm_exit(EXIT_FAILURE, "output: no dir specified: %s", args);

	/* Need to allocate before before continuing */
	if ( !spool_set_buf(s, buf_sz) )
		firestorm_exit(EXIT_FAILURE, "output: unable to allocate buffer");

	/* purely informational */
	if ( s->max_bytes )
		mesg(M_INFO,"alert: %s: max log size: %lluKB",
			s->log_dir, s->max_bytes>>10);

	/* use minutes granularity */
	if ( s->max_time ) {
		s->max_time *= 60;
		mesg(M_INFO,"alert: %s: max log age: %u hrs %u mins",
			s->log_dir,
			s->max_time/(60*60),
			(s->max_time%(60*60))/60);

		/* For scheduling periodic rotation checks */
		timeout = gcd(timeout, s->max_time);
	}

	/* say whether buffered or not */
	mesg(M_INFO, "alert: %s: %sbuffered output: %uKB buffer",
		s->log_dir,
		buf_sz ? "" : "non-",
		buf_sz/1024);

	/* ok, all done, add to the list */
	spool_link(s);
}

/* NOTE that this imposes the requirement that live captures be
 * consistent with gettimeofday. If the time changes under us, we are
 * also fucked, but that is always true so please use adjtimex! :)
 */
static void alert_timer(struct packet *pkt, int l, void *priv)
{
	struct timeval tv;

	if ( !capture_is_live() )
		return;
	
	gettimeofday(&tv, NULL);
	spool_rotate_check(&tv);
}

/* Actually open files etc.. */
void alert_conf_go(void)
{
	if ( timeout && capture_is_live() ) {
		struct itimerval t;

		t.it_interval.tv_sec = timeout;
		t.it_interval.tv_usec = 0;
		t.it_value.tv_sec = timeout;
		t.it_value.tv_usec = 0;

		if ( setitimer(ITIMER_REAL, &t, NULL) ) {
			mesg(M_ERR, "setitimer(): %u seconds: %s",
					timeout, sys_err());
			mesg(M_WARN, "Logs may not be rotated after max time");
		}else{
			mesg(M_INFO, "Using a %u second timer", timeout);
		}

		notify(NOTIFY_TIMER, alert_timer, NULL, NULL);
	}

	spool_init_all();
}
