/*
* This file is part of Firestorm NIDS
* Copyright (c) 2002,2003,2004 Gianni Tedesco
* This program is released under the terms of the GNU GPL version 2
*
* This file is the entry point of the firestorm sensor;
* it is responsible for initialising all other components.
*/
#include <firestorm.h>
#include <time.h>
#include <stdarg.h>
#include <signal.h>
#include <unistd.h>
#include <fcntl.h>

#include <mesg.h>
#include <args.h>
#include <alert.h>
#include <loader.h>
#include <signature.h>
#include <decode.h>
#include <preproc.h>
#include <capture.h>
#include <detect.h>
#include <sensor.h>

static const char default_config[]="/etc/firestorm.conf";

#ifdef HAVE_SIGACTION
static void signalfn(int signum, siginfo_t *info, void *foo)
#else
#warning "Your architecture doesn't have sigaction, uh oh..."
static void signalfn(int signum)
#endif
{
	const char *action;
	int skipmsg = 0;
	int type;

	switch(signum) {
	case SIGKILL:
	case SIGTERM:
	case SIGINT:
		action = "cleaning up";
		type = CAP_INTR_DEADLY;
		break;
	case SIGHUP:
		action = "rotating logs";
		type = CAP_INTR_HUP;
		break;
	case SIGUSR1:
		action = "user defined 1";
		type = CAP_INTR_USR1;
		break;
	case SIGUSR2:
		action = "user defined 2";
		type = CAP_INTR_USR2;
		break;
	case SIGALRM:
		action = "timer signal";
		type = CAP_INTR_TIMER;
		skipmsg = 1;
		break;
	default:
		action = "UNKNOWN";
		type = CAP_INTR_UNKNOWN;
		break;
	}

	if ( skipmsg )
		goto do_the_business;

#if 0
#ifdef HAVE_SIGACTION
	mesg(M_INFO,"signal: Got signal %u from pid=%u uid=%u: %s",
		signum, info->si_pid, info->si_uid, action);
#else
	mesg(M_INFO,"signal: Got signal %u: %s", signum, action);
#endif
#endif

do_the_business:
	capture_interrupt(type);
}

/* Called once for each packet, we run the signatures */
static void do_sigs(struct packet *p, int x, void *priv)
{
	struct layer *l;
	int i;

	/* Run signatures */
	for(i=0, l=p->layer; i<p->llen; i++,l++) {
		if ( l->proto && l->proto->sig_match )
			l->proto->sig_match(p, i);
	}

	alert(p);
}

int main(int argc, char **argv)
{
#ifdef HAVE_SIGACTION
	struct sigaction sig;
#endif

#ifdef HAVE_TZSET
	/* Set timezone from locale */
	tzset();
#endif

	/* Setup printing to screen */
	smesg_init();
	mesg_set_backend(smesg_backend);

	firestorm_init();

	/* Print copyright information */
	mesg(M_INFO,"Firestorm NIDS  v" PACKAGE_VERSION);
	mesg(M_INFO,"Copyright (c) 2002-2004 Gianni Tedesco");
	mesg(M_INFO,"This program is free software; released under "
		"the GNU GPL v2 (see: COPYING)");

	if ( argc > 2 )
		firestorm_exit(EXIT_FAILURE, "bad command line arguments");

	/* Read in all configurations */
	conf_go((argc == 2) ? argv[1] : default_config);

	if ( signature_count ) {
		mesg(M_INFO,"signature: %u signatures loaded", signature_count);
		notify(NOTIFY_PACKET, do_sigs, NULL, NULL);
	}

	if ( !detect_optimize() ) {
		firestorm_exit(EXIT_FAILURE, "Unrecoverable error during "
					"signature optimization");
	}

	/* Catch INT and TERM for clean exits */
	/* Catch HUP to rotate logs */
	/* Catch USR1 and USR2 to use later */
	/* Catch SIGALRM for itimer for logrotates */
	/* Ignore SIGPIPE (for stormwall fifo) */
#ifdef HAVE_SIGACTION
	sig.sa_sigaction=signalfn;
	sig.sa_flags=SA_SIGINFO;
	sigemptyset(&sig.sa_mask);
	sigaddset(&sig.sa_mask, SIGINT);
	sigaddset(&sig.sa_mask, SIGTERM);
	sigaddset(&sig.sa_mask, SIGUSR1);
	sigaddset(&sig.sa_mask, SIGUSR2);
	sigaddset(&sig.sa_mask, SIGALRM);
	sigaddset(&sig.sa_mask, SIGHUP);
	sigaction(SIGINT, &sig, NULL);
	sigaction(SIGTERM, &sig, NULL);
	sigaction(SIGUSR1, &sig, NULL);
	sigaction(SIGUSR2, &sig, NULL);
	sigaction(SIGALRM, &sig, NULL);
	sigaction(SIGHUP, &sig, NULL);

	sig.sa_handler=SIG_IGN;
	sig.sa_flags=SA_SIGINFO;
	sigemptyset(&sig.sa_mask);
	sigaction(SIGPIPE, &sig, NULL);
#else
	signal(SIGINT, signalfn);
	signal(SIGTERM, signalfn);
	signal(SIGUSR1, signalfn);
	signal(SIGUSR2, signalfn);
	signal(SIGALRM, signalfn);
	signal(SIGHUP, signalfn);
	signal(SIGPIPE, SIG_IGN);
#endif

	capture_go();

	firestorm_exit(EXIT_SUCCESS, "Firestorm exiting normally");
}
