/*
* This file is part of Firestorm NIDS
* Copyright (c) 2002,2003,2004 Gianni Tedesco
* This program is released under the terms of the GNU GPL version 2
*
* This little stub is the signature engine. We construct rule objects
* and pass them on to the relevent backends. The real work is then
* done inside the decode plugins. The functions here are intended for
* parser plugins. I hate middle layers.
*/
#include <firestorm.h>
#include <alert.h>
#include <signature.h>
#include <decode.h>
#include <idx.h>
#include <matcher.h>
#include <field.h>

unsigned int signature_count;

static const char *substitute(struct sub_table *s, const char *k)
{
	unsigned int i;

	if ( s == NULL )
		return k;

	for(i=0; s[i].from; i++) {
		if ( !strcmp(k, s[i].from) )
			return s[i].to;
	}

	return k;
}

static void crit_dtor(void *ptr)
{
	struct criteria *c = (struct criteria *)ptr;

	if ( c->args )
		free(c->args);
}

static void mod_dtor(void *ptr)
{
	struct modifier *c = (struct modifier *)ptr;

	if ( c->name )
		free((char *)c->name);

	if ( c->args )
		free(c->args);
}

struct rule *signature_rule(const char *proto)
{
	struct rule *ret;
	struct proto *p;

	p = decode_proto(proto);
	if ( p == NULL )
		return NULL;

	ret = calloc(1, sizeof(*ret));
	if ( ret == NULL )
		firestorm_die_error("calloc");

	if ( !mpool_new(&ret->crit, sizeof(struct criteria), 0) ) {
		free(ret);
		return NULL;
	}

	if ( !mpool_new(&ret->mod, sizeof(struct modifier), 0) ) {
		mpool_free(&ret->crit);
		free(ret);
		return NULL;
	}

	ret->protocol = p;

	return ret;
}

void signature_abort(struct rule *r)
{
	/* Target/alert stuff */
	if ( r->alert.alert )
		free(r->alert.alert);
	if ( r->alert.tag )
		free(r->alert.tag);
	mpool_destroy(&r->crit, crit_dtor);
	mpool_destroy(&r->mod, mod_dtor);
	free(r);
}

/* Add a modifier to a rule */
int signature_modifier(struct rule *r, int n, const char *c, char *val)
{
	struct criteria *crit = NULL;
	struct criteria *tmp;
	struct modifier *m;

	if ( !r->protocol || !r->crit_end )
		return 0;

	if ( r->crit_end->match->owner->allow_opts ) {
		crit = r->crit_end;
	}else{
		for(tmp=r->criteria; tmp; tmp=tmp->next) {
			if ( tmp->match->owner->allow_opts )
				crit = tmp;
		}

		if ( crit == NULL )
			return 0;
	}

	if ( !crit->match->owner->allow_opts ) {
		mesg(M_WARN, "%s: no opts allowed: (sid=%u, rev=%u)",
			crit->match->owner->name, r->alert.sid, r->alert.rev);
	}

	m = mpool_alloc_zero(&r->mod);
	if ( m == NULL )
		firestorm_die_error("mpool_zero");

	m->name = strdup(c);
	if ( m->name == NULL )
		firestorm_die_error("strdup");

	if ( val && !(m->args = strdup(val)) )
		firestorm_die_error("strdup");

	if ( n )
		m->negate = 1;
	m->next = crit->modifier;
	crit->modifier = m;

	return 1;
}

/* Add a critieria to a rule */
int signature_criteria(struct rule *r, int n, const char *c, char *val)
{
	struct criteria *crit;

	if ( !r->protocol )
		return 0;

	/* Copy all the data */
	crit = mpool_alloc_zero(&r->crit);
	if ( crit == NULL )
		firestorm_die_error("mpool_zero");

	crit->match = matcher_find(substitute(r->protocol->sig_subs, c));
	if ( crit->match == NULL ) {
		mesg(M_ERR, "sig: matcher %s (%s) not found",
			c, substitute(r->protocol->sig_subs, c));
		return 0;
	}

	if ( val && !(crit->args = strdup(val)) )
		firestorm_die_error("strdup");

	if ( n )
		crit->negate = 1;

	/* Thwack it on the end of the list */
	crit->next = NULL;
	if ( !r->criteria ) {
		r->criteria = crit;
	}else{
		r->crit_end->next = crit;
	}
	r->crit_end = crit;

	r->num_criteria++;
	return 1;
}

int signature_alert(struct rule *r,
	const char *text,
	int match_prio,
	u_int32_t sid,
	u_int32_t rev,
	unsigned long rate,
	unsigned long burst,
	u_int8_t prio)
{
	/* Set the text */
	r->alert.alert = strdup(text);
	if ( r->alert.alert == NULL )
		firestorm_die_error("strdup");

	r->alert.sid = sid;
	r->alert.rev = rev;
	r->alert.priority = prio;
	r->alert.order = match_prio;

	r->alert.t.toks = 0;
	r->alert.t.last_msg = 0;
	r->alert.t.missed = 0;
	r->alert.t.cost = rate;
	r->alert.t.burst = rate * burst;
	return 1;
}

int signature_tag(struct rule *r, struct alert_tag *tag)
{
	struct alert_tag *t;

	switch(tag->direction) {
	case TAG_DIR_SRC:
	case TAG_DIR_DST:
	case TAG_DIR_EITHER:
		if ( tag->type == TAG_TYPE_SESSION )
			goto invalid;
		break;
	case TAG_DIR_SERVER:
	case TAG_DIR_CLIENT:
		if ( tag->type == TAG_TYPE_HOST )
			goto invalid;
		break;
	default:
		break;
	}

	t = malloc(sizeof(*t));
	if ( t == NULL )
		return 0;

	memcpy(t, tag, sizeof(*t));

	mesg(M_DEBUG, "%s: tag: type=%u metric=%u count=%u dir=%u",
		r->alert.alert, t->type, t->metric, t->count, t->direction);

	r->alert.tag = t;

	return 1;
invalid:
	mesg(M_ERR, "%s: tag: tag direction invalid with tag type",
			r->alert.alert);
	return 0;
}

int signature_commit(struct rule *r)
{
	if ( !r->protocol ) {
		mesg(M_ERR, "signature: invalid signature commited");
		return 0;
	}

	if ( !r->protocol->sig_add ) {
		mesg(M_ERR, "signature: %s: protocol does not "
			"support signatures", r->protocol->name);
		return 0;
	}

	if ( r->alert.tag && !r->protocol->tag ) {
		mesg(M_ERR, "signature: %s: protocol does not "
			"support tagging", r->protocol->name);
		return 0;
	}

	if ( !r->protocol->sig_add(r) )
		return 0;

	mpool_destroy(&r->crit, crit_dtor);
	mpool_destroy(&r->mod, mod_dtor);
	free(r);
	signature_count++;
	return 1;
}
