/*
* This file is part of Firestorm NIDS
* Copyright (c) 2002,2003,2004 Gianni Tedesco
* This program is released under the terms of the GNU GPL version 2
*
* This file is the entry point of the firestorm sniffer;
* it is responsible for initialising all other components.
*/
#include <firestorm.h>
#include <time.h>
#include <stdarg.h>
#include <signal.h>
#include <unistd.h>
#include <fcntl.h>

#include <mesg.h>
#include <args.h>
#include <alert.h>
#include <loader.h>
#include <signature.h>
#include <decode.h>
#include <preproc.h>
#include <capture.h>
#include <detect.h>
#include <sensor.h>
#include <target.h>

#ifdef HAVE_GETOPT_H
#include <getopt.h>
#endif

#ifdef HAVE_SIGACTION
static void signalfn(int signum, siginfo_t *info, void *foo)
#else
#warning "Your architecture doesn't have sigaction, uh oh..."
static void signalfn(int signum)
#endif
{
	const char *action;
	int skipmsg = 0;
	int type;

	switch(signum) {
	case SIGKILL:
	case SIGTERM:
	case SIGINT:
		action = "cleaning up";
		type = CAP_INTR_DEADLY;
		break;
	case SIGHUP:
		action = "rotating logs";
		type = CAP_INTR_HUP;
		break;
	case SIGUSR1:
		action = "user defined 1";
		type = CAP_INTR_USR1;
		break;
	case SIGUSR2:
		action = "user defined 2";
		type = CAP_INTR_USR2;
		break;
	case SIGALRM:
		action = "timer signal";
		type = CAP_INTR_TIMER;
		skipmsg = 1;
		break;
	default:
		action = "UNKNOWN";
		type = CAP_INTR_UNKNOWN;
		break;
	}

	if ( skipmsg )
		goto do_the_business;

#if 0
#ifdef HAVE_SIGACTION
	mesg(M_INFO,"signal: Got signal %u from pid=%u uid=%u: %s",
		signum, info->si_pid, info->si_uid, action);
#else
	mesg(M_INFO,"signal: Got signal %u: %s", signum, action);
#endif
#endif

do_the_business:
	capture_interrupt(type);
}

/* Called once for each packet, we send to target */
static struct target *tgt;
static void *tgt_priv;
static void do_sniffer(struct packet *p, int x, void *priv)
{
	static struct generator gen = {.name = "sniffer"};
	static struct alert a = {
		.alert = "sniffer",
		.sid = 0, .rev = 0,
		.priority = 5,
		.gen = &gen,
	};
	p->alert = &a;
	tgt->target(p, tgt_priv);
}

#ifdef HAVE_GETOPT_LONG
static const struct option opts[]={
	{"plugin-dir", 	1, 0, 'p'},
	{"format", 1, 0, 'f'},
	{"options", 1, 0, 'o'},
	{"capdev", 1, 0, 'c'},
	{"source", 1, 0, 's'},
	{"verbose", 0, 0, 'v'},
	{"help", 0, 0, 'h'},
	{NULL, 0, 0, 0},
};
#endif
#define MY_OPTS "p:f:o:c:s:vh?"

static void get_plugins(void)
{
#if 1
	loader_load_dir(PLUGIN_DIR "/protocols");
	loader_load_dir(PLUGIN_DIR "/output");
#else
	loader_load_dir("decode_plugins/.libs/");
	loader_load_dir("capdev_plugins/.libs/");
	loader_load_dir("target_plugins/.libs/");
#endif
	decode_load();
	capdev_load();
	target_load();
	preproc_load();
}

static const char *cmd;
static int verbose_flag;
static void mesg_backend(unsigned char code, const char *str, size_t len)
{
	if ( !verbose_flag && code > M_WARN )
		return;

	/* Use stderr cos output may use stdout */
	fprintf(stderr, "%s: %s\n", cmd, str);
}

static void usage(int ecode) __noreturn;
static void usage(int ecode)
{
	extern struct target *targets;
	extern struct capdev *capdevs;
	struct target *t;
	struct capdev *c;

	fprintf(stderr, "Usage: %s -c [CAPDEV] -s [SOURCE] [OPTS...]\n", cmd);
	fprintf(stderr, "Packet sniffer based on the Firestorm core.\n\n");
	fprintf(stderr, "    --capdev, -c      Choose a capture type\n");
	fprintf(stderr, "    --source, -s      Capdev specific option string\n");
	fprintf(stderr, "    --format, -f      Choose an output format\n");
	fprintf(stderr, "    --options, -o     Format specific option string\n");
	fprintf(stderr, "    --plugin-dir, -p  Load extra plugins from this directory\n");
	fprintf(stderr, "    --verbose, -v     Display debugging messages to stderr\n");
	fprintf(stderr, "    --help, -h        This message\n\n");
	fprintf(stderr, "Loading plugins from: %s\n", PLUGIN_DIR "/protocols");
	fprintf(stderr, "Loading plugins from: %s\n", PLUGIN_DIR "/output");

	get_plugins();

	fprintf(stderr, "\nAvailable capdev plugins and their options:\n");
	for(c=capdevs; c; c=c->next) {
		fprintf(stderr, "  %s - %s\n", c->name, c->desc);
		if ( t->args == NULL )
			continue;
	}

	fprintf(stderr, "\nAvailable output formats and their options:\n");
	for(t=targets; t; t=t->next) {
		struct arg *a;

		fprintf(stderr, "  %s - %s\n", t->name, t->desc);
		if ( t->args == NULL )
			continue;

		for(a = t->args; a->name; a++)
			fprintf(stderr, "   o %s: %s (%s)\n",
				a->name,
				a->help ? a->help : "[no help available]",
				args_typename(a->type));
	}

	fprintf(stderr, "\nReport bugs to <gianni@scaramanga.co.uk>\n");
	firestorm_exit(ecode, NULL);
}

int main(int argc, char **argv)
{
	const char *t = "ascii";
	const char *t_opts = NULL;
	const char *cd = "tcpdump";
	const char *cd_opts = NULL;
	int c;
#ifdef HAVE_SIGACTION
	struct sigaction sig;
#endif

#ifdef HAVE_TZSET
	/* Set timezone from locale */
	tzset();
#endif
	cmd = argv[0];

	/* Setup printing to screen */
	mesg_set_backend(mesg_backend);

	firestorm_init();

	/* Print copyright information */
	mesg(M_INFO,"Firestorm Sniffer v" PACKAGE_VERSION);
	mesg(M_INFO,"Copyright (c) 2002-2004 Gianni Tedesco");
	mesg(M_INFO,"This program is free software; released under "
		"the GNU GPL v2 (see: COPYING)");

#if HAVE_GETOPT_LONG
	while ( (c=getopt_long(argc, argv, MY_OPTS, opts, NULL)) != -1 )
#else
	while ( (c=getopt(argc, argv, MY_OPTS)) != -1 )
#endif
	{
		switch ( c ) {
		case 'f':
			t = optarg;
			break;
		case 'o':
			t_opts = optarg;
			break;
		case 'c':
			cd = optarg;
			break;
		case 's':
			cd_opts = optarg;
			break;
		case 'p':
			loader_load_dir(optarg);
			break;
		case 'v':
			verbose_flag = 1;
			break;
		case 'h':
			usage(EXIT_SUCCESS);
		default:
			usage(EXIT_FAILURE);
		}
	}

	get_plugins();

	/* Use some basic preprocessors */
	if ( !(preproc_setup("ipfrag",
			"mem_hi=4096k mem_lo=3072k minttl=0 timeout=30") &&
		preproc_setup("tcpstream",
			"num_streams=32k num_flows=32k reassemble=yes") && 
		preproc_setup("http_normalize",
			"emulate=IIS")) ) {
		mesg(M_DEBUG, "preprocessors failed to load");
	}

	capture_setup(cd, cd_opts);

	tgt = target_find(t);
	if ( tgt == NULL )
		firestorm_exit(EXIT_FAILURE,
				"%s: Failed to find capdev plugin",
				t);

	if ( !tgt->validate(t_opts, &tgt_priv) )
		firestorm_exit(EXIT_FAILURE,
				"%s: Failed to initialize output: %s",
				t, t_opts);

	notify(NOTIFY_PACKET, do_sniffer, NULL, NULL);

	/* Catch INT and TERM for clean exits */
	/* Catch HUP to rotate logs */
	/* Catch USR1 and USR2 to use later */
	/* Catch SIGALRM for itimer for logrotates */
	/* Ignore SIGPIPE (for stormwall fifo) */
#ifdef HAVE_SIGACTION
	sig.sa_sigaction=signalfn;
	sig.sa_flags=SA_SIGINFO;
	sigemptyset(&sig.sa_mask);
	sigaddset(&sig.sa_mask, SIGINT);
	sigaddset(&sig.sa_mask, SIGTERM);
	sigaddset(&sig.sa_mask, SIGUSR1);
	sigaddset(&sig.sa_mask, SIGUSR2);
	sigaddset(&sig.sa_mask, SIGALRM);
	sigaddset(&sig.sa_mask, SIGHUP);
	sigaction(SIGINT, &sig, NULL);
	sigaction(SIGTERM, &sig, NULL);
	sigaction(SIGUSR1, &sig, NULL);
	sigaction(SIGUSR2, &sig, NULL);
	sigaction(SIGALRM, &sig, NULL);
	sigaction(SIGHUP, &sig, NULL);

	sig.sa_handler=SIG_IGN;
	sig.sa_flags=SA_SIGINFO;
	sigemptyset(&sig.sa_mask);
	sigaction(SIGPIPE, &sig, NULL);
#else
	signal(SIGINT, signalfn);
	signal(SIGTERM, signalfn);
	signal(SIGUSR1, signalfn);
	signal(SIGUSR2, signalfn);
	signal(SIGALRM, signalfn);
	signal(SIGHUP, signalfn);
	signal(SIGPIPE, SIG_IGN);
#endif

	capture_go();

	firestorm_exit(EXIT_SUCCESS, "Firestorm exiting normally");
}
