/*
 * This file is part of Firestorm NIDS
 * Copyright (c) 2002,2003,2004 Gianni Tedesco <gianni@scaramanga.co.uk>
 * This program is released under the terms of the GNU GPL version 2
*/
#include <firestorm.h>
#include <dotscara/fobuf.h>
#include <fcntl.h>
#include <unistd.h>
#include <time.h>
#include <stdarg.h>

#include <strtouint.h>
#include <args.h>
#include <alert.h>
#include <signature.h>
#include <decode.h>
#include <target.h>
#include <plugin.h>
#include <capture.h>

/* Private data structure */
struct ascii_priv {
	struct fobuf buf;
	int hex;
	unsigned int llen;
};

/* Arguments */
static struct arg ascii_args[]={
	{"file", ARGTYPE_FOBUF, NULL, .help = "Output filename"},
	{"hex", ARGTYPE_PBOOL, NULL, .help = "Enable hex dump"},
	{"len", ARGTYPE_PBYTES, NULL, .help = "Hex dump line length"},
	{NULL, ARGTYPE_NOP, NULL},
};

static int print_error;
static char *abuf;
static size_t abuflen;

static void ascii_printf(struct fobuf *f, const char *fmt, ...)
	__printf(2,3) __nonull(1,2);
static void ascii_printf(struct fobuf *f, const char *fmt, ...)
{
	int len;
	va_list va;
	char *new;

again:
	va_start(va, fmt);

	len = vsnprintf(abuf, abuflen, fmt, va);
	if ( len < 0 ) /* bug in old glibc */
		len = 0;
	if ( (size_t)len < abuflen )
		goto done;

	new = realloc(abuf, len + 1);
	if ( new == NULL )
		goto done;

	abuf = new;
	abuflen = len + 1;
	goto again;

done:
	if ( !fobuf_write(f, abuf, (size_t)len) )
		print_error = 1;

	va_end(va);
}

static void hex_dump(struct fobuf *f, char *ptr,
			unsigned int len, unsigned int llen)
{
	unsigned int i, j;
	unsigned int line;
	unsigned char *tmp=(unsigned char *)ptr;

	for(j=0; j<len; j+=line, tmp+=line) {
		if ( j+llen > len ) {
			line=len-j;
		}else{
			line=llen;
		}

		ascii_printf(f, "%05X : ", j);

		for(i=0; i<line; i++) {
			if ( isprint(tmp[i]) ) {
				ascii_printf(f, "%c", tmp[i]);
			}else{
				ascii_printf(f, ".");
			}
		}

		for(; i<llen; i++)
			ascii_printf(f, " ");

		for(i=0; i<line; i++)
			ascii_printf(f, " %02X", tmp[i]);

		ascii_printf(f, "\n");
	}
}

static int ascii_alert(struct packet *pkt, void *priv)
{
	static char ln[256];
	struct ascii_priv *p=priv;
	struct alert *a=pkt->alert;
	struct tm *rtime;
	int i;

	if ( !pkt )
		return 0;

	print_error = 0;

	rtime=localtime(&pkt->time.tv_sec);
	strftime(ln, sizeof(ln), "%Y-%m-%d %H:%M:%S", rtime);

	/* Basic packet information */
	ascii_printf(&p->buf, "  packet: %s.%.6lu len=%u caplen=%u\n",
		ln, pkt->time.tv_usec,
		pkt->len, pkt->caplen);

	/* Alert information */
	ascii_printf(&p->buf, "   alert: [%s] %s (sid=%u.%u prio=%u)\n",
		a->gen->name, a->alert,
		a->sid, a->rev, a->priority);

	/* Print all the layers */
	for(i=0; i<pkt->llen; i++) {
		if ( !pkt->layer[i].proto ) {
			ascii_printf(&p->buf, "    data: Application "
				"layer data (%u bytes)\n",
				pkt->end-pkt->layer[i].h.raw);
			continue;
		}

		if ( pkt->layer[i].proto->print ) {
			pkt->layer[i].proto->print(pkt, i, ln, sizeof(ln));
		}else{
			sprintf(ln, "(no information)");
		}

		ascii_printf(&p->buf, "%8s: %s\n",
			pkt->layer[i].proto->name,
			ln);
	}

	/* hex dump the mofo */
	if ( p->hex ) {
		hex_dump(&p->buf, pkt->base, pkt->caplen, p->llen);
	}

	ascii_printf(&p->buf, "\n");

	if ( print_error || !fobuf_flush(&p->buf) ) {
		mesg(M_ERR, "ascii: output error: %s", sys_err2(NULL));
		return 0;
	}

	return 1;
}

static int ascii_validate(const char *args, void **ptr)
{
	struct ascii_priv *p=NULL;

	if ( !(p=calloc(1, sizeof(*p))) )
		return 0;

	p->buf.fd = -1;

	/* Defaults */
	p->llen = 16;
	p->hex = 1;

	ascii_args[0].val.fobuf = &p->buf;
	ascii_args[1].val.vp_bool = &p->hex;
	ascii_args[2].val.vp_bytes = &p->llen;

	switch ( args_parse(ascii_args, args, p) ) {
	case -1:
		mesg(M_ERR,"ascii: parse error: %s", args);
	case 0: /* fall through */
		goto err;
	default:
		break;
	}

	if ( p->buf.fd == -1 ) {
		if ( !fobuf_attach(&p->buf, STDOUT_FILENO, 0, 0) )
			goto err;
	}

	*ptr = p;
	return 1;
err:
	if ( p ) {
		if ( p->buf.fd != -1 )
			fobuf_close(&p->buf);
		free(p);
	}
	return 0;
}

static int ascii_cleanup(void *priv)
{
	struct ascii_priv *p = priv;
	int ret = 1;

	if ( !fobuf_close(&p->buf) ) {
		mesg(M_ERR, "fobuf_close(): %s", sys_err2(NULL));
		ret = 0;
	}

	free(p);
	return ret;
}

struct target firestorm_targets[]={
	{
		.name = "ascii",
		.desc = "detailed ascii text with optional hex dumps",
		.target = ascii_alert,
		.validate = ascii_validate,
		.cleanup = ascii_cleanup,
		.args = ascii_args,
	},
	{.name = NULL},
};

struct plugin_desc firestorm_plugin_desc = {
	.size		= sizeof(firestorm_plugin_desc),
	.name		= "target.ascii",
	.desc		= "Detailed ASCII logs",
	.author_name	= "Gianni Tedesco",
	.author_email	= "gianni@scaramanga.co.uk",
	.ver_major	= 1,
	.ver_minor	= 1,
};
