/*
 * This file is part of Firestorm NIDS
 * Copyright (c) 2002,2003,2004 Gianni Tedesco <gianni@scaramanga.co.uk>
 * This program is released under the terms of the GNU GPL version 2
*/
#include <firestorm.h>
#include <dotscara/fobuf.h>
#include <fcntl.h>
#include <errno.h>
#include <unistd.h>
#include <stdarg.h>
#include <time.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <arpa/inet.h>

#include <args.h>
#include <alert.h>
#include <signature.h>
#include <decode.h>
#include <target.h>
#include <plugin.h>
#include <capture.h>

static const char * const state_str[]={
	"CLOSED",
	"ESTABLISHED",
	"SYN_SENT",
	"SYN_RECV",
	"FIN_WAIT1",
	"FIN_WAIT2",
	"TCP_TIME_WAIT",
	"TCP_CLOSE",
	"TCP_CLOSE_WAIT",
	"TCP_LAST_ACK",
	"TCP_LISTEN",
	"TCP_CLOSING"
};

static struct proto *ipproto;
static struct proto *http_proto;

static char buf[8192];
static char *pb;
static size_t pbsz;

static void print_ready(void)
{
	pb = buf;
	pbsz = sizeof(buf);
}

static int print_fini(struct fobuf *f)
{
	size_t len = sizeof(buf) - pbsz;

	if ( !fobuf_write(f, buf, len) )
		return 0;
	if ( !fobuf_flush(f) )
		return 0;
	return 1;
}

static void print_append(const char *fmt, ...) __nonull(1) __printf(1,2);
static void print_append(const char *fmt, ...)
{
	int ret;
	va_list va;

	if ( pbsz == 0 )
		return;

	va_start(va, fmt);
	ret = vsnprintf(pb, pbsz, fmt, va);
	va_end(va);

	/* bug in old glibc */
	if ( ret < 0 )
		ret = 0;

	if ( (size_t)ret >= pbsz ) {
		pbsz = 0;
		pb = buf + sizeof(buf);
	}else{
		pbsz -= (size_t)ret;
		pb += (size_t)ret;
	}
}

static void log_http(struct http_request *h)
{
	struct fvec *uri;
	size_t sz;

	if ( h->type != HTTP_REQUEST )
		return;

	if ( h->uri_path.ptr ) {
		uri = &h->uri_path;
	}else{
		uri = &h->uri;
	}

	sz = 27 + uri->len + h->method.len;

	/* 28 = 14+12+2 = extraneous bits of string */
	if ( pbsz < sz )
		return;

	pbsz -= sz;

	memcpy(pb, " http_method='", 14);
	pb += 14;

	memcpy(pb, h->method.ptr, h->method.len);
	pb += h->method.len;

	memcpy(pb, "' http_uri='", 12);
	pb += 12;

	memcpy(pb, uri->ptr, uri->len);
	pb += uri->len;

	memcpy(pb, "'", 2);
	pb += 1;
}

static int log_alert(struct packet *pkt, void *priv)
{
	union {
		struct pkt_tcphdr *tcph;
		struct pkt_udphdr *udph;
		struct pkt_icmphdr *icmph;
	}u;
	u_int32_t flags;
	struct alert *a = pkt->alert;
	struct fobuf *f = priv;
	struct pkt_iphdr *iph;
	char sip[16], dip[16];
	struct in_addr ina;
	int l;

	/* TCP stuff */
	char tcpflags[]="FSRPAU12";
	char disp[9];
	int n, x;
	struct tcp_session *s;
	int s_state, c_state;

	if ( !pkt )
		return 0;

	for(l=0; l<pkt->llen; l++) {
		if ( pkt->layer[l].proto==ipproto )
			goto found;
	}

	mesg(M_CRIT,"log: No IP header found");
	return 1;

found:
	print_ready();

	/* Check we have an ip header */
	if ( pkt->layer[l].h.raw + sizeof(*iph) > pkt->end )
		return 1;

	iph=pkt->layer[l].h.ip;

	/* Convert IPs to strings */
	ina.s_addr=iph->saddr;
	snprintf(sip, sizeof(sip), "%s", inet_ntoa(ina));
	ina.s_addr=iph->daddr;
	snprintf(dip, sizeof(dip), "%s", inet_ntoa(ina));

	print_append("%lu.%.6lu alert=%s sig=%u.%u "
		"priority=%u src=%s dst=%s proto=%u ",
		pkt->time.tv_sec,
		pkt->time.tv_usec,
		a->gen->name,
		a->sid, a->rev,
		a->priority,
		sip, dip,
		iph->protocol & 0xff);

	if ( l+1 >= pkt->llen )
		goto just_ip;

	u.tcph=pkt->layer[l+1].h.tcp;
	flags=pkt->layer[l+1].flags;

	switch(iph->protocol) {
	case 6:
		for(n=0,x=1; n<8; n++, x*=2) {
			if ( u.tcph->flags.flags & x ){
				disp[n]=tcpflags[n];
			}else{
				disp[n]='*';
			}
		}disp[8]='\0';

		if ( (s=pkt->layer[l+1].session) ) {
			s_state=s->server.state;
			c_state=s->client.state;
		}else{
			s_state=0;
			c_state=0;
		}

		print_append("spt=%u dpt=%u flags=%s "
			"from=%s server=%s client=%s",
			ntohs(u.tcph->sport)&0xffff,
			ntohs(u.tcph->dport)&0xffff,
			disp,
			(flags&FLAG_TCP_2SVR) ? "client" : "server",
			state_str[s_state],
			state_str[c_state]);

		if ( l+2<pkt->llen &&
			pkt->layer[l+2].proto==http_proto
			&& pkt->layer[l+2].session ) {
			log_http(pkt->layer[l+2].session);
		}

		break;
	case 17:
		print_append("spt=%u dpt=%u",
			ntohs(u.udph->sport)&0xffff,
			ntohs(u.udph->dport)&0xffff);
		break;
	case 1:
		print_append("type=%u code=%u",
			u.icmph->type&0xff,
			u.icmph->code&0xff);
		break;
	default:
		break;
	}

just_ip:
	print_append(" : %s\n", a->alert);
	return print_fini(f);

}

static struct arg log_args[]={
	{"file", ARGTYPE_FOBUF, NULL, .help = "Output filename"},
	{NULL, ARGTYPE_NOP, NULL},
};

static int log_validate(const char *args, void **ptr)
{
	struct fobuf *f;

	if ( !ipproto && !(ipproto=decode_proto("ip")) ) {
		mesg(M_ERR,"log: IP protocol must be loaded "
			"to use log output");
		return 0;
	}

	if ( !http_proto && !(http_proto=decode_proto("http")) ) {
		mesg(M_ERR,"log: HTTP protocol must be loaded "
			"to use log output");
		return 0;
	}

	f = calloc(1, sizeof(*f));
	if ( f == NULL ) {
		mesg(M_ERR, "calloc(): %s", sys_err());
		goto err;
	}

	f->fd = -1;
	log_args[0].val.fobuf = f;

	switch ( args_parse(log_args, args, NULL) ) {
	case -1:
		mesg(M_ERR,"log: parse error: %s", args);
	case 0: /* fall through */
		goto err;
	default:
		break;
	}

	if ( f->fd == -1 ) {
		if ( !fobuf_attach(f, STDOUT_FILENO, 0, 0) ) {
			mesg(M_ERR, "fobuf_attach(): %s", sys_err2(NULL));
			goto err;
		}
	}

	*ptr = f;
	return 1;
err:
	if ( f ) {
		if ( f->fd != -1 )
			fobuf_close(f);
		free(f);
	}

	return 0;
}

static int log_cleanup(void *priv)
{
	struct fobuf *f = priv;
	int ret = 1;

	if ( !fobuf_close(f) ) {
		mesg(M_ERR, "fobuf_close(): %s", sys_err2(NULL));
		ret = 0;
	}

	free(f);

	return ret;
}

struct target firestorm_targets[]={
	{
		.name = "log",
		.desc = "one line per packet textual format",
		.target = log_alert,
		.validate = log_validate,
		.cleanup = log_cleanup,
		.args = log_args,
	},
#if 0
	{
		.name = "syslog",
		.desc = "sends messages to syslog",
		.target = log_alert,
		.validate = log_validate,
		.cleanup = log_cleanup,
		.args = syslog_args,
	}
#endif
	{.name = NULL}
};

struct plugin_desc firestorm_plugin_desc = {
	.size		= sizeof(firestorm_plugin_desc),
	.name		= "target.log",
	.desc		= "TCP/IP based log plugin",
	.author_name	= "Gianni Tedesco",
	.author_email	= "gianni@scaramanga.co.uk",
	.ver_major	= 0,
	.ver_minor	= 1,
};
